import subprocess
import threading

from .test_utils import (
    BaseManagerTestCase,
)

from pulsar.managers.unqueued import CoexecutionManager


class Coexecutor:
    """Mimic shell script in other container of coexecutor pod-like environment."""

    def __init__(self, manager):
        self.manager = manager
        self.has_command_line = False
        self.command_line = None

    def monitor(self):
        while not self.has_command_line:
            try:
                command_line = self.manager.read_command_line("123")
            except (OSError, ValueError):
                continue
            if not command_line:
                # might be partially written... need to be make this atomic I think.
                continue
            self.command_line = command_line
            self.has_command_line = True

        subprocess.call(command_line, shell=True)
        # we are ignoring this exit code and just trusting the one in the job script...
        # I'm not sure what to do about that.


class CoexecutionManagerTest(BaseManagerTestCase):

    def setUp(self):
        super().setUp()
        self._set_manager()

    def tearDown(self):
        super().setUp()

    def _set_manager(self, **kwds):
        self.manager = CoexecutionManager('_default_', self.app, **kwds)

    def test_simple_execution(self):
        coexecutor = Coexecutor(self.manager)
        t = threading.Thread(target=coexecutor.monitor)
        t.start()
        try:
            self._test_simple_execution(self.manager, timeout=5)
        finally:
            coexecutor.has_command_line = True
            t.join(2)



from .test_utils import (
    BaseManagerTestCase,
    skip_unless_module
)

from pulsar.managers.queued_drmaa import DrmaaQueueManager


class DrmaaManagerTest(BaseManagerTestCase):

    def setUp(self):
        super().setUp()
        self._set_manager()

    def tearDown(self):
        super().setUp()
        self.manager.shutdown()

    def _set_manager(self, **kwds):
        self.manager = DrmaaQueueManager('_default_', self.app, **kwds)

    @skip_unless_module("drmaa")
    def test_simple_execution(self):
        self._test_simple_execution(self.manager)

    @skip_unless_module("drmaa")
    def test_cancel(self):
        self._test_cancelling(self.manager)



"""Script used to run an example job against a running Pulsar server.

Exercises various features both the Pulsar client and server.
"""

import logging
import optparse
import os
import re
import shutil
import sys
import tempfile
import threading
import time
import traceback
from collections import namedtuple

from galaxy.tool_util.deps.dependencies import DependenciesDescription
from galaxy.tool_util.deps.requirements import ToolRequirement

from pulsar.client import (
    build_client_manager,
    CLIENT_INPUT_PATH_TYPES,
    ClientInput,
    ClientInputs,
    ClientJobDescription,
    ClientOutputs,
    finish_job,
    PulsarOutputs,
    submit_job,
)
from .test_common import write_config

TEST_SCRIPT = b"""# -*- coding: utf-8 -*-
from __future__ import print_function

import sys

from os import getenv
from os import listdir
from os import makedirs
from os.path import abspath
from os.path import basename
from os.path import dirname
from os.path import exists
from os.path import join

SCRIPT_DIRECTORY = abspath(dirname(__file__))
HELPER_DIRECTORY = join(SCRIPT_DIRECTORY, "subdir")

print("stdout output")
def assert_path_contents(path, expected_contents):
    if not exists(path):
        message = "Expected path [%s] to exist, but it doesn't."
        raise AssertionError(message % path)

    with open(path, 'r') as f:
        contents = f.read()
        if contents != expected_contents:
            message = "Expected path [%s] to be have contents [%s], but contains [%s]."
            raise AssertionError(message % (path, expected_contents, contents))

config_input = open(sys.argv[1], 'r')
output = open(sys.argv[3], 'w')
output2 = open(sys.argv[5], 'w')
output2_contents = sys.argv[6]
output3 = open(sys.argv[7], 'w')
version_output = open(sys.argv[9], 'w')
index_path = sys.argv[10]
assert len(listdir(dirname(index_path))) == 2
assert len(listdir(join(dirname(dirname(index_path)), "seq"))) == 1
output4_index_path = open(sys.argv[11], 'w')
metadata_dir = dirname(sys.argv[13])
output5 = open(sys.argv[15], 'w')
legacy_galaxy_json = int(sys.argv[16])
output_metadata_path = join(metadata_dir, "metadata_output")
tmp_dir_exists_path = join(".", "tmp_dir_test")
home_dir_exists_path = join(".", "home_dir_test")
try:
    assert_path_contents(sys.argv[2], "Hello world input!!@!")
    assert_path_contents(sys.argv[8], "INPUT_EXTRA_CONTENTS")
    assert_path_contents(sys.argv[13], "meta input")
    assert_path_contents(sys.argv[14], "INPUT METADATA CONTENTS...")
    contents = config_input.read(1024)
    output.write(contents)
    open("workdir_output", "w").write("WORK DIR OUTPUT")
    open("env_test", "w").write(getenv("TEST_ENV", "DEFAULT"))
    open("rewrite_action_test", "w").write(sys.argv[12])
    output2.write(output2_contents)
    if legacy_galaxy_json:
        with open("galaxy.json", "w") as f: f.write('''{"filename": "gjson_refer"}\n{"filename": "gjson_refer2"}''')
    else:
        with open("galaxy.json", "w") as f: f.write('''[{"filename": "gjson_refer"}, {"filename": "gjson_refer2"}]''')

    with open("gjson_refer", "w") as f: f.write('''gjson_refer_contents''')
    with open("gjson_refer2", "w") as f: f.write('''gjson_refer_contents2''')
    with open(output_metadata_path, "w") as f: f.write("meta output")
    tmp_dir = getenv("_GALAXY_JOB_TMP_DIR", "fakepathnotreal")
    tmp_dir_exists = str(exists(tmp_dir))
    with open(tmp_dir_exists_path, "w") as f: f.write(f"tmp-{tmp_dir_exists}")
    home_dir = getenv("_GALAXY_JOB_HOME_DIR", "fakepathnotreal")
    home_dir_exists = str(exists(home_dir))
    with open(home_dir_exists_path, "w") as f: f.write(f"home-{home_dir_exists}")

    output3.write(getenv("MOO", "moo_default"))
    output1_extras_path = "%s_files" % sys.argv[3][0:-len(".dat")]
    makedirs(output1_extras_path)
    open(join(output1_extras_path, "extra"), "w").write("EXTRA_OUTPUT_CONTENTS")
    version_output.write("1.0.1")
    output4_index_path.write(index_path)
    output5.write(str(exists(join(HELPER_DIRECTORY, "helper.R"))))
finally:
    output.close()
    config_input.close()
    output2.close()
    output3.close()
    version_output.close()
    output4_index_path.close()
sys.exit(4)
"""

HELP_URL = "URL of the Pulsar web server to target."
HELP_PRIVATE_TOKEN = ("Private token used to authorize client, if the "
                      "Pulsar server specified a private_token in app.yml "
                      "this must match that value.")
HELP_TRANSPORT = "Specify as 'curl' to use pycurl client for staging."
HELP_CACHE = "Specify to test Pulsar caching during staging."
HELP_TEST_ERRORS = "Specify to exercise exception handling during staging."
HELP_SUPPRESS_OUTPUT = ""
HELP_DISABLE_CLEANUP = ("Specify to disable cleanup after the job, this "
                        "is useful to checking the files generated during "
                        "the job and stored on the Pulsar server.")
HELP_JOB_ID = "Submit the Pulsar job with this 'external' id."
HELP_DEBUG = "Enable debug log output from Pulsar client"

EXPECTED_OUTPUT = b"hello world output"
EXAMPLE_UNICODE_TEXT = 'єχαмρℓє συтρυт'
TEST_REQUIREMENT = ToolRequirement(name="dep1", version="1.1", type="package")
TEST_DEPENDENCIES = DependenciesDescription(requirements=[TEST_REQUIREMENT])

ClientInfo = namedtuple("ClientInfo", ["client", "client_manager"])


class MockTool:

    def __init__(self, tool_dir):
        self.id = "client_test"
        self.version = "1.0"
        self.tool_dir = tool_dir


class TestRequiredFilesObject:
    # it would be nice to use a RequiredFiles but that hasn't been published to PyPI yet, so just
    # stick to the interface here.

    def find_required_files(self, tool_directory):
        return ["script.py", "subdir/helper.R"]


def run(options):
    logging.basicConfig(level=logging.DEBUG)
    waiter = None
    client_manager = None
    try:
        temp_directory = tempfile.mkdtemp(prefix='pulsar-check-client.')
        temp_index_dir = os.path.join(temp_directory, "idx", "bwa")
        temp_index_dir_sibbling = os.path.join(temp_directory, "idx", "seq")
        temp_shared_dir = os.path.join(temp_directory, "shared", "test1")
        temp_work_dir = os.path.join(temp_directory, "w")
        temp_metadata_dir = os.path.join(temp_directory, "m")
        temp_false_working_dir = os.path.join(temp_metadata_dir, "working")
        temp_tool_dir = os.path.join(temp_directory, "t")
        temp_tool_sub_dir = os.path.join(temp_tool_dir, "subdir")

        __makedirs([
            temp_tool_dir,
            temp_tool_sub_dir,
            temp_work_dir,
            temp_index_dir,
            temp_index_dir_sibbling,
            temp_shared_dir,
            temp_metadata_dir,
            temp_false_working_dir,
        ])

        temp_input_path = os.path.join(temp_directory, "dataset_0.dat")
        temp_input_extra_path = os.path.join(temp_directory, "dataset_0_files", "input_subdir", "extra")
        temp_input_metadata_path = os.path.join(temp_directory, "metadata", "12312231231231.dat")
        temp_index_path = os.path.join(temp_index_dir, "human.fa")

        temp_config_path = os.path.join(temp_work_dir, "config.txt")
        temp_tool_path = os.path.join(temp_directory, "t", "script.py")
        temp_tool_helper_path = os.path.join(temp_directory, "t", "subdir", "helper.R")
        temp_output_path = os.path.join(temp_directory, "dataset_1.dat")
        temp_output2_path = os.path.join(temp_directory, "dataset_2.dat")
        temp_output3_path = os.path.join(temp_directory, "dataset_3.dat")
        temp_output4_path = os.path.join(temp_directory, "dataset_4.dat")
        temp_output5_path = os.path.join(temp_directory, "dataset_5.dat")
        temp_version_output_path = os.path.join(temp_directory, "GALAXY_VERSION_1234")
        temp_output_workdir_destination = os.path.join(temp_directory, "dataset_77.dat")
        temp_output_workdir = os.path.join(temp_work_dir, "env_test")
        temp_metadata_path = os.path.join(temp_metadata_dir, "metadata_test123")

        temp_output_workdir_destination2 = os.path.join(temp_directory, "dataset_78.dat")
        temp_output_workdir2 = os.path.join(temp_work_dir, "rewrite_action_test")

        temp_output_workdir_destination3 = os.path.join(temp_directory, "dataset_79.dat")
        temp_output_workdir3 = os.path.join(temp_work_dir, "tmp_dir_test")
        temp_output_workdir_destination4 = os.path.join(temp_directory, "dataset_80.dat")
        temp_output_workdir4 = os.path.join(temp_work_dir, "home_dir_test")

        __write_to_file(temp_input_path, b"Hello world input!!@!")
        __write_to_file(temp_input_extra_path, b"INPUT_EXTRA_CONTENTS")
        __write_to_file(temp_input_metadata_path, b"INPUT METADATA CONTENTS...")
        __write_to_file(temp_config_path, EXPECTED_OUTPUT)
        __write_to_file(temp_metadata_path, "meta input")
        __write_to_file(temp_tool_path, TEST_SCRIPT)
        __write_to_file(temp_tool_helper_path, b"helper R file")
        __write_to_file(temp_index_path, b"AGTC")
        # Implicit files that should also get transferred since depth > 0
        __write_to_file("%s.fai" % temp_index_path, b"AGTC")
        __write_to_file(os.path.join(temp_index_dir_sibbling, "human_full_seqs"), b"AGTC")

        empty_input = "/foo/bar/x"

        test_unicode = getattr(options, "test_unicode", False)  # TODO Switch this in integration tests
        legacy_galaxy_json = getattr(options, "legacy_galaxy_json", False)
        cmd_text = EXAMPLE_UNICODE_TEXT if test_unicode else "Hello World"
        command_line_params = (
            temp_tool_path,
            temp_config_path,
            temp_input_path,
            temp_output_path,
            empty_input,
            temp_output2_path,
            cmd_text,
            temp_output3_path,
            temp_input_extra_path,
            temp_version_output_path,
            temp_index_path,
            temp_output4_path,
            temp_shared_dir,
            temp_metadata_path,
            temp_input_metadata_path,
            temp_output5_path,
            "1" if legacy_galaxy_json else "0",
        )
        assert os.path.exists(temp_index_path)
        command_line = 'python %s "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s"' % command_line_params
        config_files = [temp_config_path]
        client_inputs = []
        client_inputs.append(ClientInput(temp_input_path, CLIENT_INPUT_PATH_TYPES.INPUT_PATH))
        client_inputs.append(ClientInput(temp_input_path, CLIENT_INPUT_PATH_TYPES.INPUT_PATH))
        # Reverting empty input handling added in:
        #  https://github.com/galaxyproject/pulsar/commit/2fb36ba979cf047a595c53cdef833cae79cbb380
        # Seems like it really should cause a failure.
        # client_inputs.append(ClientInput(empty_input, CLIENT_INPUT_PATH_TYPES.INPUT_PATH))
        client_inputs.append(ClientInput(os.path.join(temp_directory, "dataset_0_files"), CLIENT_INPUT_PATH_TYPES.INPUT_EXTRA_FILES_PATH))
        client_inputs.append(ClientInput(temp_input_metadata_path, CLIENT_INPUT_PATH_TYPES.INPUT_METADATA_PATH))
        output_files = [
            temp_output_path,
            temp_output2_path,
            temp_output3_path,
            temp_output4_path,
            temp_output5_path,
            temp_output_workdir_destination,
            temp_output_workdir_destination2,
            temp_output_workdir_destination3,
            temp_output_workdir_destination4,
        ]
        client, client_manager = __client(temp_directory, options)
        waiter = Waiter(client, client_manager)
        galaxy_json_type = "legacy_galaxy" if legacy_galaxy_json else "galaxy"
        dynamic_file_sources = [
            {"type": galaxy_json_type, "path": "galaxy.json"},
            {"type": galaxy_json_type, "path": "absent_galaxy.json"},
        ]
        client_outputs = ClientOutputs(
            working_directory=temp_work_dir,
            metadata_directory=temp_metadata_dir,
            work_dir_outputs=[
                (temp_output_workdir, temp_output_workdir_destination),
                (temp_output_workdir2, temp_output_workdir_destination2),
                (temp_output_workdir3, temp_output_workdir_destination3),
                (temp_output_workdir4, temp_output_workdir_destination4),
            ],
            output_files=output_files,
            version_file=temp_version_output_path,
            dynamic_file_sources=dynamic_file_sources,
        )
        job_description = ClientJobDescription(
            command_line=command_line,
            tool=MockTool(temp_tool_dir),
            config_files=config_files,
            client_inputs=ClientInputs(client_inputs),
            client_outputs=client_outputs,
            working_directory=temp_work_dir,
            metadata_directory=temp_metadata_dir,
            **__extra_job_description_kwargs(options)
        )
        job_id = submit_job(client, job_description)
        if job_id:
            client.job_id = job_id
        result_status = waiter.wait()

        expecting_full_metadata = getattr(options, "expecting_full_metadata", True)
        if expecting_full_metadata:
            stdout = result_status["stdout"].strip()
            std_streams_debug = f"actual stdout [{stdout}], actual stderr [{result_status['stderr']}]"
            assert "stdout output".startswith(stdout), f"Standard output is not an initial substring of [stdout output], {std_streams_debug}"

            if hasattr(options, "maximum_stream_size"):
                assert len(stdout) == options.maximum_stream_size

        assert result_status["complete"] == "true"
        if expecting_full_metadata:
            assert result_status["returncode"] == 4, f"Expected exit code of 4, got {result_status['returncode']} - {std_streams_debug}"
        if expecting_full_metadata:
            __finish(options, client, client_outputs, result_status)
        else:
            client.clean()
        __assert_contents(temp_output_path, EXPECTED_OUTPUT, result_status)
        __assert_contents(temp_output2_path, cmd_text, result_status)
        if not legacy_galaxy_json:
            expected_contents = b'''[{"filename": "gjson_refer"}, {"filename": "gjson_refer2"}]'''
            __assert_contents(os.path.join(temp_work_dir, "galaxy.json"), expected_contents, result_status)
        __assert_contents(os.path.join(temp_work_dir, "gjson_refer"), b"gjson_refer_contents", result_status)
        __assert_contents(os.path.join(temp_work_dir, "gjson_refer2"), b"gjson_refer_contents2", result_status)
        __assert_contents(os.path.join(temp_directory, "dataset_1_files", "extra"), b"EXTRA_OUTPUT_CONTENTS", result_status)
        __assert_contents(os.path.join(temp_metadata_dir, "metadata_output"), b"meta output", result_status)
        if getattr(options, "test_rewrite_action", False):
            __assert_contents(temp_output_workdir_destination2, os.path.join(temp_directory, "shared2", "test1"), result_status)
        __assert_contents(temp_output_workdir_destination3, "tmp-True", result_status)
        __assert_contents(temp_output_workdir_destination4, "home-True", result_status)
        if job_description.env:
            __assert_contents(temp_output_workdir_destination, b"TEST_ENV_VALUE", result_status)
        __assert_contents(temp_version_output_path, b"1.0.1", result_status)
        if job_description.dependencies_description:
            __assert_contents(temp_output3_path, "moo_override", result_status)
        else:
            __assert_contents(temp_output3_path, "moo_default", result_status)
        client_job_id = getattr(options, "job_id", "123456")
        # Revise this test so it works for containers... or fix whatever is broken.
        __assert_has_rewritten_bwa_path(client, client_job_id, temp_output4_path)
        if getattr(options, "explicit_tool_declarations", False):
            __assert_contents(temp_output5_path, str(True), result_status)
        __exercise_errors(options, client, temp_output_path, temp_directory)
    except BaseException:
        if not options.suppress_output:
            traceback.print_exc()
        raise
    finally:
        if client_manager:
            client_manager.shutdown()
        if waiter is not None:
            waiter.shutdown()
        if getattr(options, 'cleanup', True):
            shutil.rmtree(temp_directory)


class Waiter:

    def __init__(self, client, client_manager):
        self.client = client
        self.client_manager = client_manager
        self.background = hasattr(client_manager, 'ensure_has_status_update_callback')
        self.__setup_callback()

    def __setup_callback(self):
        if self.background:
            self.event = threading.Event()

            def on_update(message):
                if message["status"] in ["complete", "cancelled"]:
                    self.final_status = message
                    self.event.set()

            self.client_manager.ensure_has_status_update_callback(on_update)

    def wait(self, seconds=120):
        final_status = None
        if not self.background:
            i = 0
            # Wait for seconds * 2 half second intervals
            while i < (seconds * 2):
                complete_response = self.client.raw_check_complete()
                if complete_response["status"] in ["complete", "cancelled"]:
                    final_status = complete_response
                    break
                time.sleep(.5)
                i = i + 1
        else:
            self.event.wait(seconds)
            if self.event.is_set():
                final_status = self.final_status
        if not final_status:
            raise Exception("Job not completed properly")

        return final_status

    def shutdown(self):
        client_manager = self.client_manager
        client_manager.shutdown(ensure_cleanup=True)


def __assert_contents(path, expected_contents, pulsar_state):
    if not os.path.exists(path):
        raise AssertionError("File {} not created. Final Pulsar response state [{}]".format(path, pulsar_state))
    if isinstance(expected_contents, bytes):
        file = open(path, 'rb')
    else:
        file = open(path, encoding="utf-8")
    try:
        contents = file.read()
        if contents != expected_contents:
            message = "File ({}) contained invalid contents [{}].".format(path, contents)
            message = "{} Expected contents [{}]. Final Pulsar response state [{}]".format(message, expected_contents, pulsar_state)
            raise AssertionError(message)
    finally:
        file.close()


def __assert_has_rewritten_bwa_path(client, job_id, temp_output4_path):
    if client.default_file_action != "none" and job_id:
        rewritten_index_path = open(temp_output4_path, encoding='utf-8').read()
        # Path written to this file will differ between Windows and Linux.
        patt = r"%s[/\\]unstructured[/\\]\w+[/\\]bwa[/\\]human.fa" % job_id
        if re.search(patt, rewritten_index_path) is None:
            raise AssertionError("[%s] does not contain rewritten path [%s]." % (patt, rewritten_index_path))


def __exercise_errors(options, client, temp_output_path, temp_directory):
    """
    Exercise error conditions.

    TODO: Improve. Something should be checked here.
    """
    if getattr(options, 'test_errors', False):
        try:
            client._fetch_output(temp_output_path + "x")
        except BaseException:
            if not options.suppress_output:
                traceback.print_exc()


def __client(temp_directory, options):
    client_options = extract_client_options(options)
    default_file_action = client_options.get("default_file_action", None)
    unstructured_action = default_file_action or "transfer"
    path_defs = [
        dict(path=os.path.join(temp_directory, "idx"), path_types="unstructured", depth=2, action=unstructured_action),
    ]
    if getattr(options, "test_rewrite_action", False):
        rewrite_def = dict(
            path=os.path.join(temp_directory, "shared"),
            path_types="unstructured",
            action="rewrite",
            source_directory=os.path.join(temp_directory, "shared"),
            destination_directory=os.path.join(temp_directory, "shared2")
        )
        path_defs.append(rewrite_def)
    client_options["file_action_config"] = write_config(temp_directory, dict(paths=path_defs))
    if default_file_action in ["remote_scp_transfer", "remote_rsync_transfer"]:
        test_key = os.environ["PULSAR_TEST_KEY"]
        if not test_key.startswith("----"):
            test_key = open(test_key).read()
        client_options["ssh_key"] = test_key
        client_options["ssh_user"] = os.environ.get("USER")
        client_options["ssh_port"] = 22
        client_options["ssh_host"] = "localhost"

    user = getattr(options, 'user', None)
    if user:
        client_options["submit_user"] = user
    job_id = getattr(options, "job_id", "123456")
    return client_info(options, client_options, job_id=job_id)


def extract_client_options(options):
    """ Exract options explicitly related to build client from
    configured client manager.

    """
    default_file_action = getattr(options, "default_file_action", None)
    client_options = {
        "url": getattr(options, "url", None),
        "private_token": getattr(options, "private_token", None),
    }
    if default_file_action:
        client_options["default_file_action"] = default_file_action
    if hasattr(options, "jobs_directory"):
        client_options["jobs_directory"] = options.jobs_directory
    if hasattr(options, "files_endpoint"):
        client_options["files_endpoint"] = options.files_endpoint
    if hasattr(options, "k8s_enabled"):
        client_options["k8s_enabled"] = options.k8s_enabled
    if hasattr(options, "tes_url"):
        client_options["tes_url"] = options.tes_url
    if hasattr(options, "container"):
        client_options["container"] = options.container
    return client_options


def client_info(options, client_options, job_id=None):
    """ From command-line arguments ``options`` - extract options
    related to build a client manager and build it. Then get a client
    with supplied client options and optional job id.
    """
    if job_id is None:
        job_id = options.job_id
    client_manager = client_manager_from_args(options)
    client = client_manager.get_client(client_options, job_id)
    return ClientInfo(client, client_manager)


def client_manager_from_args(options):
    manager_args = {}
    simple_client_manager_options = ['cache', 'job_manager', 'file_cache']
    for client_manager_option in simple_client_manager_options:
        if getattr(options, client_manager_option, None):
            manager_args[client_manager_option] = getattr(options, client_manager_option)
    if getattr(options, 'transport', None):
        manager_args['transport'] = options.transport
    if getattr(options, 'manager_url', None):
        manager_args['amqp_url'] = options.manager_url
    if getattr(options, 'tes_url', None):
        manager_args['tes_url'] = options.tes_url
    if getattr(options, "k8s_enabled", None):
        manager_args['k8s_enabled'] = options.k8s_enabled
    cm = build_client_manager(**manager_args)
    return cm


def __write_to_file(path, contents):
    dirname = os.path.dirname(path)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    with open(path, "wb") as f:
        if not isinstance(contents, bytes):
            contents = bytes(contents, "UTF-8")
        f.write(contents)


def __makedirs(directories):
    for directory in directories:
        os.makedirs(directory)


def __extra_job_description_kwargs(options):
    dependencies_description = None
    test_requirement = getattr(options, "test_requirement", False)
    if test_requirement:
        requirements = [TEST_REQUIREMENT]
        dependencies_description = DependenciesDescription(requirements=requirements)
    test_env = getattr(options, "test_env", False)
    env = []
    if test_env:
        env.append(dict(name="TEST_ENV", value="TEST_ENV_VALUE"))
    container = getattr(options, "container", None)
    remote_pulsar_app_config = getattr(options, "remote_pulsar_app_config", None)
    rval = dict(dependencies_description=dependencies_description, env=env, container=container, remote_pulsar_app_config=remote_pulsar_app_config)
    if getattr(options, "explicit_tool_declarations", False):
        rval["tool_directory_required_files"] = TestRequiredFilesObject()
    return rval


def __finish(options, client, client_outputs, result_status):
    pulsar_outputs = PulsarOutputs.from_status_response(result_status)
    cleanup_job = 'always'
    if not getattr(options, 'cleanup', True):
        cleanup_job = 'never'
    finish_args = dict(
        client=client,
        job_completed_normally=True,
        cleanup_job=cleanup_job,  # Default should 'always' if overridden via options.
        client_outputs=client_outputs,
        pulsar_outputs=pulsar_outputs,
    )
    failed = finish_job(**finish_args)
    if failed:
        failed_message_template = "Failed to complete job correctly, final status %s, finish exceptions %s."
        failed_message = failed_message_template % (result_status, failed)
        raise AssertionError(failed_message)


def main(argv=None):
    """ Exercises a running Pulsar with the Pulsar client. """
    mod_docstring = sys.modules[__name__].__doc__
    parser = optparse.OptionParser(mod_docstring)
    parser.add_option('--url', dest='url', default='http://localhost:8913/', help=HELP_URL)
    parser.add_option('--private_token', default=None, help=HELP_PRIVATE_TOKEN)
    parser.add_option('--transport', default=None, choices=["urllib", "curl"], help=HELP_TRANSPORT)  # set to curl to use pycurl
    parser.add_option('--cache', default=False, action="store_true", help=HELP_CACHE)
    parser.add_option('--test_errors', default=False, action="store_true", help=HELP_TEST_ERRORS)
    parser.add_option('--suppress_output', default=False, action="store_true", help=HELP_SUPPRESS_OUTPUT)
    parser.add_option('--disable_cleanup', dest="cleanup", default=True, action="store_false", help=HELP_DISABLE_CLEANUP)
    parser.add_option('--job_id', default="123456", help=HELP_JOB_ID)
    parser.add_option('--explicit_tool_declarations', default=False, action="store_true")
    parser.add_option('--legacy_galaxy_json', default=False, action="store_true")
    parser.add_option('--debug', default=False, action="store_true", help=HELP_DEBUG)
    (options, args) = parser.parse_args(argv)
    run(options)


if __name__ == "__main__":
    main()



import os
from .test_utils import (
    TempDirectoryTestCase,
    test_pulsar_server,
)
from pulsar.scripts import run


class ScriptRunTestCase(TempDirectoryTestCase):

    def setUp(self):
        super().setUp()
        input1 = os.path.join(self._working_directory, "input1")
        open(input1, "w").write("Hello World!")
        self.input1 = input1

    def simple_test(self):
        app_conf = {}
        with test_pulsar_server(app_conf=app_conf) as server:
            url = server.application_url
            self._run([
                "--url", url,
                "--default_file_action", "transfer",
            ])
            self._check_outputs()

    def _run(self, pulsar_args):
        run_args = pulsar_args[:]
        run_args.extend([
            "--command", "echo `pwd` > output1; cp input1 output_test2",
            "--working_directory", self._working_directory,
            "--output", "output1",
            "--output_pattern", r"output_test\d",
            "--result_json", self._result,
        ])
        exit_code = run.main(run_args)
        if os.path.exists(self._result):
            print(open(self._result).read())
        else:
            assert False, "No result json file found"
        assert exit_code == 0

    def _check_outputs(self):
        output1 = os.path.join(self._working_directory, "output1")
        output_test2 = os.path.join(self._working_directory, "output_test2")
        # Prove it went somewhere else :)
        assert open(output1).read() != self._working_directory
        assert open(output_test2).read() == "Hello World!"

    @property
    def _result(self):
        return os.path.join(self.temp_directory, "r.json")

    @property
    def _working_directory(self):
        work_dir = os.path.join(self.temp_directory, "work")
        if not os.path.exists(work_dir):
            os.makedirs(work_dir)
        return work_dir



from contextlib import contextmanager
import os.path

from pulsar import manager_factory

from .test_utils import temp_directory
from .test_utils import minimal_app_for_managers


def test_default():
    # By defualt create single queue python manager with name _default_
    with __test_managers({}) as managers:
        assert len(managers) == 1
        assert "_default_" in managers
        __assert_manager_of_type(managers["_default_"], "queued_python")


def test_ini_populate():
    with temp_directory() as dir:
        config = os.path.join(dir, "job_managers.ini")
        open(config, "w").write("[manager:cool1]\ntype=queued_cli")
        with __test_managers({"job_managers_config": config}) as managers:
            assert len(managers) == 1
            assert "cool1" in managers
            __assert_manager_of_type(managers["cool1"], "queued_cli")


def test_yaml_populate_manager():
    conf = {
        'manager': {'type': 'queued_cli'}
    }
    with __test_managers(conf) as managers:
        assert len(managers) == 1
        assert "_default_" in managers
        __assert_manager_of_type(managers["_default_"], "queued_cli")


def test_yaml_populate_managers():
    conf = {
        'managers': {
            'man1': {'type': 'queued_cli'},
            'man2': {'type': 'queued_python'}
        },
    }

    with __test_managers(conf) as managers:
        assert len(managers) == 2
        assert "man1" in managers
        __assert_manager_of_type(managers["man1"], "queued_cli")
        __assert_manager_of_type(managers["man2"], "queued_python")


def __assert_manager_of_type(manager, expected_type):
    # Not a great test - break stateful abstraction and is testing
    # implementation details instead of behavior.
    actual = manager._proxied_manager.manager_type
    assert actual == expected_type


@contextmanager
def __test_managers(conf):
    app = minimal_app_for_managers()
    managers = manager_factory.build_managers(app, conf)
    try:
        yield managers
    finally:
        for manager in managers.values():
            manager.shutdown()



import contextlib
import os
import threading
import time

from .test_utils import (
    TempDirectoryTestCase,
    skip_unless_module,
    skip_without_drmaa,
    restartable_pulsar_app_provider,
    integration_test,
)
from pulsar.manager_endpoint_util import (
    submit_job,
)
from pulsar.managers.stateful import ActiveJobs
from pulsar.client.amqp_exchange import ACK_FORCE_NOACK_KEY
from pulsar.client.amqp_exchange_factory import get_exchange
from pulsar.managers.util.drmaa import DrmaaSessionFactory


class StateIntegrationTestCase(TempDirectoryTestCase):

    @skip_without_drmaa
    @skip_unless_module("kombu")
    @integration_test
    def test_restart_finishes_job(self):
        test = "restart_finishes"
        with self._setup_app_provider(test) as app_provider:
            job_id = '12345'

            with app_provider.new_app() as app:
                manager = app.only_manager
                job_info = {
                    'job_id': job_id,
                    'command_line': 'sleep 1000',
                    'setup': True,
                }
                submit_job(manager, job_info)
                external_id = None
                for i in range(10):
                    time.sleep(.05)
                    # TODO: unfortunate breaking of abstractions here.
                    external_id = manager._proxied_manager._external_id(job_id)
                    if external_id:
                        break
                if external_id is None:
                    assert False, "Test failed, couldn't get exteranl id for job id."

            drmaa_session = DrmaaSessionFactory().get()
            drmaa_session.kill(external_id)
            drmaa_session.close()
            consumer = self._status_update_consumer(test)
            consumer.start()

            with app_provider.new_app() as app:
                consumer.wait_for_messages()

            consumer.join()
            assert len(consumer.messages) == 1, len(consumer.messages)
            assert consumer.messages[0]["status"] == "complete"

    @skip_unless_module("drmaa")
    @skip_unless_module("kombu")
    @integration_test
    def test_recovery_failure_fires_lost_status(self):
        test = "restart_failure_fires_lost"
        with self._setup_app_provider(test) as app_provider:
            job_id = '12345'

            with app_provider.new_app() as app:
                persistence_directory = app.persistence_directory

            # Break some abstractions to activate a job that
            # never existed.
            manager_name = "manager_%s" % test
            active_jobs = ActiveJobs(manager_name, persistence_directory)
            active_jobs.activate_job(job_id)

            consumer = self._status_update_consumer(test)
            consumer.start()

            with app_provider.new_app() as app:
                consumer.wait_for_messages()

            consumer.join()

            assert len(consumer.messages) == 1, len(consumer.messages)
            assert consumer.messages[0]["status"] == "lost"

    @skip_unless_module("kombu")
    @integration_test
    def test_staging_failure_fires_failed_status(self):
        test = "stating_failure_fires_failed"
        with self._setup_app_provider(test, manager_type="queued_python") as app_provider:
            job_id = '12345'

            consumer = self._status_update_consumer(test)
            consumer.start()

            with app_provider.new_app() as app:
                manager = app.only_manager
                job_info = {
                    'job_id': job_id,
                    'command_line': 'sleep 1000',
                    'setup': True,
                    # Invalid staging description...
                    'remote_staging': {"setup": [{"moo": "cow"}]}
                }
                # TODO: redo this with submit_job coming through MQ for test consistency.
                submit_job(manager, job_info)

            import time
            time.sleep(2)
            consumer.wait_for_messages()
            consumer.join()

            assert len(consumer.messages) == 1, len(consumer.messages)
            assert consumer.messages[0]["status"] == "failed"

    @skip_unless_module("kombu")
    @integration_test
    def test_async_request_of_mq_status(self):
        test = "async_request_of_mq_status"
        with self._setup_app_provider(test, manager_type="queued_python") as app_provider:
            job_id = '12345'

            consumer = self._status_update_consumer(test)
            consumer.start()

            with app_provider.new_app() as app:
                manager = app.only_manager
                job_info = {
                    'job_id': job_id,
                    'command_line': 'sleep 1000',
                    'setup': True,
                    # Invalid staging description...
                    'remote_staging': {"setup": [{"moo": "cow"}]}
                }
                # TODO: redo this with submit_job coming through MQ for test consistency.
                submit_job(manager, job_info)
                self._request_status(test, job_id)

            import time
            time.sleep(2)
            consumer.wait_for_messages()
            consumer.join()

            messages = consumer.messages
            assert len(messages) == 2, len(messages)
            assert messages[0]["status"] == "failed"
            assert messages[1]["status"] == "failed", messages[1]

    @skip_unless_module("kombu")
    @integration_test
    def test_async_request_of_mq_status_lost(self):
        test = "async_request_of_mq_status_lost"
        with self._setup_app_provider(test, manager_type="queued_python") as app_provider:
            job_id = '12347'  # should be lost? - never existed right?

            consumer = self._status_update_consumer(test)
            consumer.start()

            with app_provider.new_app() as app:
                app.only_manager
                # do two messages to ensure generation of status message doesn't
                # create a job directory we don't mean to or something like that
                self._request_status(test, job_id)
                self._request_status(test, job_id)

            import time
            time.sleep(2)
            consumer.wait_for_messages()
            consumer.join()

            messages = consumer.messages
            assert len(messages) == 2, len(messages)
            assert messages[0]["status"] == "lost", messages[0]
            assert messages[1]["status"] == "lost", messages[1]

    @skip_unless_module("kombu")
    @integration_test
    def test_setup_failure_fires_failed_status(self):
        test = "stating_failure_fires_failed"
        with self._setup_app_provider(test, manager_type="queued_python") as app_provider:
            job_id = '12345'

            consumer = self._status_update_consumer(test)
            consumer.start()

            with app_provider.new_app() as app:
                manager = app.only_manager
                job_info = {
                    'job_id': job_id,
                    'command_line': 'sleep 1000',
                    'setup': True,
                }

                with open(os.path.join(app_provider.staging_directory, job_id), "w") as f:
                    f.write("File where staging directory should be, setup should fail now.")

                # TODO: redo this with submit_job coming through MQ for test consistency,
                # would eliminate the need for the exception catch as well.
                try:
                    submit_job(manager, job_info)
                except Exception:
                    pass

            consumer.wait_for_messages()
            consumer.join()

            assert len(consumer.messages) == 1, len(consumer.messages)
            assert consumer.messages[0]["status"] == "failed"

    @contextlib.contextmanager
    def _setup_app_provider(self, test, manager_type="queued_drmaa"):
        mq_url = "memory://test_%s" % test
        manager = "manager_%s" % test
        app_conf = dict(message_queue_url=mq_url)
        app_conf["managers"] = {manager: {'type': manager_type}}
        with restartable_pulsar_app_provider(app_conf=app_conf, web=False) as app_provider:
            yield app_provider

    def _status_update_consumer(self, test):
        mq_url = "memory://test_%s" % test
        manager = "manager_%s" % test
        consumer = SimpleConsumer(queue="status_update", url=mq_url, manager=manager)
        return consumer

    def _request_status(self, test, job_id):
        mq_url = "memory://test_%s" % test
        manager = "manager_%s" % test
        exchange = get_exchange(mq_url, manager, {})
        params = {
            "job_id": job_id,
            ACK_FORCE_NOACK_KEY: True,
        }
        exchange.publish("status", params)


class SimpleConsumer:

    def __init__(self, queue, url, manager="_default_"):
        self.queue = queue
        self.url = url
        self.manager = manager
        self.active = True
        self.exchange = get_exchange(url, manager, {})

        self.messages = []

    def start(self):
        t = threading.Thread(target=self._run)
        t.start()
        self.thread = t

    def join(self):
        self.active = False
        self.thread.join(10)

    def wait_for_messages(self, n=1):
        accumulate_time = 0.0
        while len(self.messages) < n:
            time.sleep(.1)
            accumulate_time += 0.05
            if accumulate_time > 3.0:
                raise Exception("Waited too long for messages.")

    def _run(self):
        self.exchange.consume("status_update", self._callback, check=self)

    def _callback(self, body, message):
        self.messages.append(body)
        message.ack()

    def __nonzero__(self):
        return self.active

    __bool__ = __nonzero__  # Both needed Py2 v 3



""" Test utilities in pulsar.main """
from os.path import join
from .test_utils import temp_directory
from pulsar import main


def test_pulsar_config_builder_defaults():
    with temp_directory() as mock_root:
        __write_mock_ini(join(mock_root, "server.ini"))
        config = main.PulsarConfigBuilder(config_dir=mock_root)
        assert config.load()["foo"] == "bar1"


def test_pulsar_config_builder_defaults_sample():
    with temp_directory() as mock_root:
        __write_mock_ini(join(mock_root, "server.ini.sample"))
        config = main.PulsarConfigBuilder(config_dir=mock_root)
        assert config.load()["foo"] == "bar1"


def test_pulsar_config_builder_specified_ini():
    with temp_directory() as mock_root:
        __write_mock_ini(join(mock_root, "moo.ini"))
        config = main.PulsarConfigBuilder(config_dir=mock_root, ini_path="moo.ini")
        assert config.load()["foo"] == "bar1"


def test_pulsar_config_builder_specified_ini_args():
    with temp_directory() as mock_root:
        __write_mock_ini(join(mock_root, "moo.ini"), app="cool1")
        config = main.PulsarConfigBuilder(config_dir=mock_root, args=MockArgs("moo.ini", "cool1"))
        assert config.load()["foo"] == "bar1"


def test_pulsar_config_builder_specified_app():
    with temp_directory() as mock_root:
        __write_mock_ini(join(mock_root, "server.ini"), app="cool1")
        config = main.PulsarConfigBuilder(config_dir=mock_root, app="cool1")
        assert config.load()["foo"] == "bar1"


def test_pulsar_config_builder_app_yaml():
    with temp_directory() as mock_root:
        __write_mock_ini(join(mock_root, "server.ini"))
        open(join(mock_root, "app.yml"), "w").write("foo: bar2")
        config = main.PulsarConfigBuilder(config_dir=mock_root)
        assert config.load()["foo"] == "bar2"


def test_pulsar_config_builder_override_app_yaml():
    with temp_directory() as mock_root:
        app_yaml_path = join(mock_root, "new_app.yml")
        __write_mock_ini(join(mock_root, "server.ini"), extra="app_config=%s" % app_yaml_path)
        open(app_yaml_path, "w").write("foo: bar2")
        config = main.PulsarConfigBuilder(config_dir=mock_root)
        assert config.load()["foo"] == "bar2"


def test_pulsar_manager_config_builder_defaults():
    with temp_directory() as mock_root:
        __write_mock_ini(join(mock_root, "server.ini"))
        config = main.PulsarManagerConfigBuilder(config_dir=mock_root)
        config.load()["foo"] == "bar1"
        as_dict = config.to_dict()
        assert as_dict["manager"] == "_default_"
        assert as_dict["ini_path"] == join(mock_root, "server.ini")
        assert as_dict["app"] == "main"


def test_pulsar_manager_config_builder_overrides():
    with temp_directory() as mock_root:
        __write_mock_ini(join(mock_root, "pulsar5.ini"), app="cool1")
        config = main.PulsarManagerConfigBuilder(config_dir=mock_root, app="cool1", ini_path="pulsar5.ini", manager="manager3")
        config.load()["foo"] == "bar1"
        as_dict = config.to_dict()
        assert as_dict["manager"] == "manager3"
        assert as_dict["ini_path"] == join(mock_root, "pulsar5.ini")
        assert as_dict["app"] == "cool1"


class MockArgs:

    def __init__(self, ini_path, app):
        self.ini_path = ini_path
        self.app_conf_path = None
        self.app_conf_base64 = None
        self.app = app


def __write_mock_ini(path, **kwds):
    contents = __mock_ini_contents(**kwds)
    open(path, "w").write(contents)


def __mock_ini_contents(app="main", extra=""):
    return """
[app:{}]
paste.app_factory = pulsar.web.wsgi:app_factory
foo=bar1
{}
""".format(app, extra)



from os.path import join

from pulsar.managers.base import JobDirectory
from pulsar.tools.validator import ExpressionValidator
from .test_utils import TempDirectoryTestCase


class ValidatorTest(TempDirectoryTestCase):

    def test_literal(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
        </command_validator>"""
        self.__assertValid(xml, "tophat2")
        self.__assertInvalid(xml, "bowtie")

    def test_two_literals(self):
        xml = """
        <command_validator>
            <literal value="python" />
            <literal value="setup.py" />
        </command_validator>"""
        self.__assertValid(xml, "python setup.py")
        self.__assertInvalid(xml, "pythonsetup.py")

    def test_parameter(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <parameter name="--mate-std-dev">
                <literal value="4" />
            </parameter>
        </command_validator>"""
        self.__assertValid(xml, "tophat2 --mate-std-dev 4")
        self.__assertValid(xml, "tophat2 --mate-std-dev=4")
        self.__assertInvalid(xml, "tophat2 --mate-std-dev=5")

    def test_integer(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <parameter name="--mate-std-dev">
                <integer />
            </parameter>
        </command_validator>"""
        self.__assertValid(xml, "tophat2 --mate-std-dev 4")
        self.__assertValid(xml, "tophat2 --mate-std-dev=4")
        self.__assertInvalid(xml, "tophat2 --mate-std-dev=5.0")

    def test_float(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <parameter name="--mate-std-dev">
                <float />
            </parameter>
        </command_validator>"""
        self.__assertValid(xml, "tophat2 --mate-std-dev 4")
        self.__assertValid(xml, "tophat2 --mate-std-dev=4")
        self.__assertValid(xml, "tophat2 --mate-std-dev=5.0")
        self.__assertValid(xml, "tophat2 --mate-std-dev=4e10")
        self.__assertValid(xml, "tophat2 --mate-std-dev=-1.0e10")
        self.__assertValid(xml, "tophat2 --mate-std-dev=-.0e10")
        self.__assertInvalid(xml, "tophat2 --mate-std-dev=cat")

    def test_tool_wrapper(self):
        xml = """
        <command_validator>
            <tool_wrapper name="tool1_wrapper.py" />
        </command_validator>
        """
        self.__assertValid(xml, "%s" % self.__job_file('tool_files', 'tool1_wrapper.py'))
        self.__assertInvalid(xml, "tool1_wrapper.py")

    def test_config_file(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <configfile name="top_opts" />
        </command_validator>
        """
        self.__assertValid(xml, "tophat2 %s" % self.__job_file('configs', 'top_opts'))
        self.__assertInvalid(xml, "tophat2 ../%s" % self.__job_file('configs', 'top_opts'))
        self.__assertInvalid(xml, "tophat2 %s" % self.__job_file('configs', 'top_optsX'))

    def test_input_file(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <input />
        </command_validator>
        """
        self.__assertValid(xml, "tophat2 %s" % self.__job_file("inputs", "dataset_23412.dat"))
        self.__assertInvalid(xml, "tophat2 %s/../../../dataset23412.dat" % self.__job_file("inputs", "dataset_23412.dat"))
        self.__assertInvalid(xml, "tophat2 ../%s" % self.__job_file("inputs", "dataset_23412.dat"))

    def test_two_inputs(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <input />
            <input />
        </command_validator>
        """
        self.__assertValid(xml, "tophat2 {} {}".format(
            self.__job_file("inputs", "dataset_23412.dat"),
            self.__job_file("inputs", "dataset_1.dat")
        ))

        self.__assertInvalid(xml, "tophat2 {} ../{}".format(
            self.__job_file("inputs", "dataset_23412.dat"),
            self.__job_file("inputs", "dataset_1.dat")
        ))

    def test_outputs_file(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <output />
            <output />
        </command_validator>
        """
        self.__assertValid(xml, "tophat2 {} {}".format(
            self.__job_file("outputs", "dataset_23412.dat"),
            self.__job_file("outputs", "dataset_1.dat")
        ))

        self.__assertInvalid(xml, "tophat2 {} ../{}".format(
            self.__job_file("outputs", "dataset_23412.dat"),
            self.__job_file("outputs", "dataset_1.dat")
        ))

    def test_outputs_from_work_dir(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <output />
            <output from_work_dir="junctions.bed" />
        </command_validator>
        """
        self.__assertValid(xml, "tophat2 {} {}".format(
            self.__job_file("outputs", "dataset_23412.dat"),
            self.__job_file("working", "junctions.bed")
        ))

        self.__assertInvalid(xml, "tophat2 {} ../{}".format(
            self.__job_file("outputs", "dataset_23412.dat"),
            self.__job_file("working", "..", "junctions.bed")
        ))

    def test_single_quotes(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <parameter name="--mate-std-dev">
                <literal value="4" single_quote="true" />
            </parameter>
        </command_validator>"""
        self.__assertValid(xml, "tophat2 --mate-std-dev '4'")
        self.__assertValid(xml, "tophat2 --mate-std-dev='4'")
        self.__assertInvalid(xml, "tophat2 --mate-std-dev=4")
        self.__assertInvalid(xml, "tophat2 --mate-std-dev=\"4\"")

    def test_double_quotes(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <parameter name="--mate-std-dev">
                <literal value="4" double_quote="true" />
            </parameter>
        </command_validator>"""
        self.__assertValid(xml, "tophat2 --mate-std-dev \"4\"")
        self.__assertValid(xml, "tophat2 --mate-std-dev=\"4\"")
        self.__assertInvalid(xml, "tophat2 --mate-std-dev=4")
        self.__assertInvalid(xml, "tophat2 --mate-std-dev='4'")

    def test_min(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <parameter name="--mate-std-dev" min="0">
                <literal value="4" double_quote="true" />
            </parameter>
        </command_validator>"""
        self.__assertValid(xml, "tophat2 --mate-std-dev \"4\"")
        self.__assertValid(xml, "tophat2 ")
        self.__assertInvalid(xml, "tophat2 --mate-std-dev=5")

    def test_max(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <regex value="[a-z]" max="2"/>
        </command_validator>"""
        self.__assertValid(xml, "tophat2 a")
        self.__assertValid(xml, "tophat2 a b")
        self.__assertInvalid(xml, "tophat2 a b c")

    def test_group(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <group>
                <literal value="a" />
                <regex value="[b-z]+" />
                <literal value="a" />
            </group>
        </command_validator>"""
        self.__assertValid(xml, "tophat2 aba")
        self.__assertValid(xml, "tophat2 abba")
        self.__assertInvalid(xml, "tophat2 abbbaa")
        self.__assertInvalid(xml, "tophat2 abb")

    def test_group_separate_by(self):
        xml = """
        <command_validator>
            <literal value="tophat2" />
            <group separate_by="x">
                <literal value="a" />
                <regex value="[b-z]+" />
                <literal value="a" />
            </group>
        </command_validator>"""
        self.__assertValid(xml, "tophat2 axbxa")
        self.__assertValid(xml, "tophat2 axbxbxa")
        self.__assertInvalid(xml, "tophat2 abba")
        self.__assertInvalid(xml, "tophat2 axbxbxbxaxa")

    def __job_file(self, *args):
        return join(self.temp_directory, '1', *args)

    def __validator(self, xml):
        return ExpressionValidator(xml)

    @property
    def job_directory(self):
        return JobDirectory(self.temp_directory, '1')

    def __is_valid(self, xml, contents):
        return self.__validator(xml).validate(self.job_directory, contents)

    def __assertValid(self, xml, contents):
        self.assertTrue(self.__is_valid(xml, contents), "{} did not validate against {}".format(contents, xml))

    def __assertInvalid(self, xml, contents):
        self.assertFalse(self.__is_valid(xml, contents), "{} falsely validated against {}".format(contents, xml))



#!/usr/bin/env python

import os
import subprocess
from string import Template


COMMAND_TEMPLATE = Template('''
${header}
======================================

${command_help}
''')

COMMANDS_TEMPLATE = """
Scripts
-----------------

This section describes some of the various scripts that are distributed with
Pulsar.

.. include:: scripts/pulsar.rst
"""

scripts = {
    "pulsar_config_windows": "pulsar-config (Windows)",
    "pulsar_config": "pulsar-config (\\*nix)",
    "pulsar_check": "pulsar-check",
    "pulsar_main": "pulsar-main"
}

command_doc_dir = os.path.join("docs", "scripts")
commands = COMMANDS_TEMPLATE

for (section, script_description) in scripts.items():
    command = script_description.split(" ")[0]
    if "Windows" in script_description:
        os.environ["MOCK_WINDOWS"] = "1"
    elif "MOCK_WINDOWS" in os.environ:
        del os.environ["MOCK_WINDOWS"]

    output = subprocess.check_output([command, "--help"])

    # def clean_rst_line(line):
    #     if line.startswith("    "):
    #         return line[4:]
    #     else:
    #         return line

    lines = output.split("\n")
    new_lines = []
    help_lines = False
    option_lines = False
    for line in lines:
        if line.lower().startswith("usage: "):
            new_lines.append("**Usage**::\n\n    %s" % line[len("usage: "):])
            usage_lines = True
        elif usage_lines and line.startswith(" "):
            new_lines.append("    %s" % line[len("usage: "):])
        elif usage_lines:
            usage_lines = False
            new_lines.append("\n**Help**\n")
            help_lines = True
        elif line.startswith("Options:"):
            # optparse-style
            help_lines = False
            new_lines.append("**Options**::\n\n")
            option_lines = True
        elif line.startswith("optional arguments:"):
            # argparse-style
            help_lines = False
            new_lines.append("**Options**::\n\n")
            option_lines = True
        elif option_lines:
            new_lines.append("    %s" % line)
        elif help_lines:
            new_lines.append(line)
    header = script_description.split(" ")
    header[0] = "``%s``" % header[0]
    text = COMMAND_TEMPLATE.safe_substitute(
        header=" ".join(header),
        command_help="\n".join(new_lines),
    )
    commands += "\n.. include:: scripts/%s.rst" % section
    open(os.path.join(command_doc_dir, section + ".rst"), "w").write(text)

open(os.path.join("docs", "scripts.rst"), "w").write(commands)



import collections
import configparser
import os
import subprocess
import sys
from io import StringIO

import yaml


from pulsar.scripts import config
from pulsar.scripts.config import main
from .test_utils import (
    skip_unless_environ,
    temp_directory,
)


def test_default_web_config():
    with temp_directory() as project_dir:
        main(["--directory", project_dir])
        project = _check_project_directory(project_dir)
        assert project.ini_config is not None

        local_env = os.path.join(project_dir, "local_env.sh")
        assert os.path.exists(local_env)
        exit_code = subprocess.check_call(['/bin/bash', '-c', '. %s' % local_env])
        assert exit_code == 0


def test_private_token():
    with temp_directory() as project_dir:
        main(["--directory", project_dir, "--private_token", "moo"])
        project = _check_project_directory(project_dir)
        assert project.ini_config is not None
        assert "private_token" in project.app_config
        assert project.app_config["private_token"] == "moo"


def test_windows_options_limited():
    help = _get_help(mock_windows=True)
    assert "--libdrmaa_path" not in help


def test_linux_options_full():
    help = _get_help(mock_windows=False)
    assert "--libdrmaa_path" in help


def test_mq_config():
    with temp_directory() as project_dir:
        main(["--directory", project_dir, "--mq"])
        project = _check_project_directory(project_dir)
        assert project.ini_config is None
        assert "message_queue_url" in project.app_config


@skip_unless_environ('USER')
def test_with_supervisor():
    with temp_directory() as project_dir:
        main(["--directory", project_dir, "--supervisor"])
        project = _check_project_directory(project_dir)
        assert project.ini_config is not None

        supervisor_conf = os.path.join(project_dir, "supervisor.conf")
        assert os.path.exists(supervisor_conf)


def test_libdrmaa_config():
    with temp_directory() as project_dir:
        real_pip = config.pip
        config.pip = MockPip()
        try:
            main(["--directory", project_dir, "--libdrmaa_path", "/path/to/test/libdrmaa.so", "--install"])

            local_env = os.path.join(project_dir, "local_env.sh")
            assert os.path.exists(local_env)
            exit_code = subprocess.check_call(['/bin/bash', '-c', '. %s' % local_env])
            assert exit_code == 0

            pip_calls = config.pip.main_calls
            assert len(pip_calls) == 1
            assert pip_calls[0] == (["install", "drmaa"],), pip_calls
        finally:
            config.pip = real_pip


def test_force():
    with temp_directory() as project_dir:
        # Write a default configuration and make sure mq is configured.
        main(["--directory", project_dir])
        project = _check_project_directory(project_dir)
        assert "message_queue_url" not in (project.app_config or {})

        # Try to re-config with message queue, expect error because files
        # already exist.
        exit_code = None
        try:
            main(["--directory", project_dir, "--mq"])
        except SystemExit as e:
            exit_code = e.code
        assert exit_code == 1

        # Try re-config again with --force, expect it to work and for MQ to be
        # configured.
        main(["--directory", project_dir, "--mq", "--force"])
        project = _check_project_directory(project_dir)
        assert "message_queue_url" in project.app_config


def _check_project_directory(project_dir):
    def path_if_exists(name):
        path = os.path.join(project_dir, name)
        if os.path.exists(path):
            return path
        else:
            return None

    app_config = None
    app_config_path = path_if_exists("app.yml")
    if app_config_path:
        app_config = yaml.safe_load(open(app_config_path))
        assert isinstance(app_config, dict) or (app_config is None)

    ini_config = None
    ini_path = path_if_exists("server.ini")
    if ini_path:
        ini_config = configparser.ConfigParser()
        ini_config.read([ini_path])

    return Project(ini_config, app_config)


Project = collections.namedtuple('Project', ['ini_config', 'app_config'])


class MockPip:

    def __init__(self):
        self.main_calls = []

    def main(self, *args):
        self.main_calls.append(args)


def _get_help(mock_windows=False):
    is_windows = config.IS_WINDOWS
    backup = sys.stdout
    out = StringIO()
    sys.stdout = out
    try:
        config.IS_WINDOWS = mock_windows
        try:
            main(["--help"])
        except SystemExit:
            pass
    finally:
        config.IS_WINDOWS = is_windows
        sys.stdout = backup

    help = out.getvalue()
    return help



from pulsar.managers.util.cli import factory


def test_torque_cli():
    job_params = dict(
        plugin="Torque",
        Priority="4",
    )
    job = __build_job_interface(job_params)

    command = job.submit("/tmp/path/test.sh")
    assert command == "qsub /tmp/path/test.sh"

    command = job.delete("5")
    assert command == "qdel 5"

    command = job.get_status()
    assert command == "qstat -x"

    job_script_kwargs = job.job_script_kwargs("/tmp/1/o", "/tmp/1/e", "my_job")
    headers = job_script_kwargs["headers"]
    # Verify priority from job params pass through
    assert "#PBS -p 4" in headers

    # Verify it just skips invalid status...
    assert job.parse_status("Warning: Network Proble.", ["24", "25"]) is None

    output = "<Data><Job><Job_Id>24</Job_Id><job_state>Q</job_state></Job>"
    output += "<Job><Job_Id>25</Job_Id><job_state>E</job_state></Job>"
    output += "<Job><Job_Id>26</Job_Id><job_state>E</job_state></Job></Data>\n"
    status_dict = job.parse_status(output, ["24", "25"])
    assert status_dict["24"] == "queued"
    assert status_dict["25"] == "running"
    assert "26" not in status_dict


def test_slurm_cli():
    job_params = dict(
        plugin="Slurm",
        ncpus=5
    )
    job = __build_job_interface(job_params)

    command = job.submit("/tmp/path/test.sh")
    assert command == "sbatch /tmp/path/test.sh"

    command = job.get_status()
    assert command == '''squeue -a -o '%A %t\''''

    command = job.get_single_status("4")
    assert command == '''squeue -a -o '%A %t' -j 4'''

    job_script_kwargs = job.job_script_kwargs("/tmp/1/o", "/tmp/1/e", "my_job")
    headers = job_script_kwargs["headers"]
    assert "#SBATCH -o /tmp/1/o" in headers
    assert "#SBATCH -e /tmp/1/e" in headers
    assert "#SBATCH -J my_job" in headers
    assert "#SBATCH -c 5" in headers

    output = """JOBID ST\n24 PD\n25 R\n26 F\n"""
    status_dict = job.parse_status(output, ["24", "25"])

    assert status_dict["24"] == "queued"
    assert status_dict["25"] == "running"
    assert "26" not in status_dict

    status = job.parse_single_status("slurm_load_jobs error: Invalid job id specified", "34")
    assert status == "complete"

    status = job.parse_single_status("JOBID ST\n26 F\n", "26")
    assert status == "failed"


def test_slurm_torque():
    job_params = dict(
        plugin="SlurmTorque",
    )
    job = __build_job_interface(job_params)

    command = job.submit("/tmp/path/test.sh")
    assert command == "qsub /tmp/path/test.sh"

    command = job.get_status()
    assert command == "qstat"

    output = """Job ID              Jobname          Username        Time     S Queue          \n"""
    output += """24                  script.sh        john            00:00:00 R debug          \n"""
    output += """27                  script2.sh        mary            00:00:00 R debug          \n"""
    stats = job.parse_status(output, ["24", "25"])
    assert stats.get("24", None) == "running", stats
    assert "25" not in stats


def __build_job_interface(job_params):
    cli_interface = factory.build_cli_interface()
    _, job = cli_interface.get_plugins({}, job_params)
    return job



from os.path import join

from pulsar.web.routes import _output_path
from .test_utils import test_manager


def test_output_path():
    with test_manager() as manager:
        path = _output_path(manager, '1', 'moo', 'direct')
        assert path == join(manager.job_directory('1').outputs_directory(), 'moo')


def test_output_path_security():
    """
    Attempt to download a file outside of a valid result directory,
    ensure it fails.
    """
    with test_manager() as manager:
        raised_exception = False
        try:
            _output_path(manager, '1', '../moo', 'direct')
        except Exception:
            raised_exception = True
        assert raised_exception



from os import environ
from pathlib import Path
import platform
from typing import (
    Dict,
    Optional,
)

from galaxy.util.bunch import Bunch
from pydantictes.api import TesClient
from pydantictes.funnelfixture import funnel_client
import pytest


from pulsar.client.test.check import run
from .test_utils import (
    files_server,
    integration_test,
    IntegrationTestConfiguration,
    mark,
    skip_unless_any_module,
    skip_unless_environ,
    skip_unless_executable,
    skip_unless_module,
    skip_without_drmaa,
    test_pulsar_app,
    test_pulsar_server,
)

TEST_TOOL_CONTAINER = "conda/miniconda3"


@integration_test
def test_integration_no_requirement(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        private_token=None,
    )


@integration_test
def test_integration_maximum_stream_size(integration_test_configuration: IntegrationTestConfiguration):
    integration_test_configuration.set_app_conf_props(
        maximum_stream_size=4
    )
    run_job(
        integration_test_configuration,
        private_token=None,
        maximum_stream_size=4,
    )


@skip_without_drmaa
@integration_test
def test_integration_as_user(external_queue_test_configuration: IntegrationTestConfiguration):
    job_props = {'type': 'queued_external_drmaa', "production": "false"}
    run_job(
        external_queue_test_configuration,
        job_conf_props=job_props,
        private_token=None,
        default_file_action="copy",
        user='u1'
    )


@integration_test
def test_integration_local_setup(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        private_token=None,
        default_file_action="remote_copy",
        local_setup=True
    )


@skip_unless_module("kombu")
@integration_test
def test_message_queue(external_queue_test_configuration: IntegrationTestConfiguration):
    mq = "memory://test_basicmq_%s" % external_queue_test_configuration.test_suffix
    external_queue_test_configuration.set_app_conf_props(message_queue_url=mq)
    run_job(
        external_queue_test_configuration,
        private_token=None,
        default_file_action="remote_copy",
        local_setup=True,
        manager_url=mq,
    )


@skip_unless_environ("PULSAR_TEST_KEY")
@skip_unless_module("kombu")
@integration_test
def test_integration_scp(external_queue_test_configuration: IntegrationTestConfiguration):
    mq = "memory://test_scp_%s" % external_queue_test_configuration.test_suffix
    external_queue_test_configuration.set_app_conf_props(message_queue_url=mq)
    run_job(
        external_queue_test_configuration,
        private_token=None,
        default_file_action="remote_scp_transfer",
        local_setup=True,
        manager_url=mq,
    )


@skip_unless_environ("PULSAR_TEST_KEY")
@skip_unless_module("kombu")
@integration_test
def test_integration_rsync(external_queue_test_configuration: IntegrationTestConfiguration):
    mq = "memory://test_rsync_%s" % external_queue_test_configuration.test_suffix
    external_queue_test_configuration.set_app_conf_props(message_queue_url=mq)
    run_job(
        external_queue_test_configuration,
        private_token=None,
        default_file_action="remote_rsync_transfer",
        local_setup=True,
        manager_url=mq,
    )


@integration_test
def test_integration_copy(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        private_token=None,
        default_file_action="copy",
    )


@integration_test
def test_integration_no_transfer(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        private_token=None,
        default_file_action="none"
    )


@integration_test
def test_integration_cached(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        private_token=None,
        cache=True,
    )


@integration_test
def test_integration_legacy_galaxy_json(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        private_token=None,
        legacy_galaxy_json=True,
    )


@integration_test
def test_integration_default(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        private_token=None
    )


@skip_unless_module("pycurl")
@integration_test
def test_integration_curl(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        private_token=None,
        transport="curl"
    )


@integration_test
def test_integration_explicit_tool_directory_includes(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        private_token=None,
        explicit_tool_declarations=True
    )


@integration_test
def test_integration_token(integration_test_configuration: IntegrationTestConfiguration):
    integration_test_configuration.set_app_conf_props(private_token="testtoken")
    run_job(
        integration_test_configuration,
        private_token="testtoken"
    )


@integration_test
def test_integration_errors(integration_test_configuration: IntegrationTestConfiguration):
    integration_test_configuration.set_app_conf_props(private_token="testtoken")
    run_job(
        integration_test_configuration,
        private_token="testtoken",
        test_errors=True
    )


@skip_without_drmaa
@integration_test
def test_integration_drmaa(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        job_conf_props={'type': 'queued_drmaa'},
        private_token=None
    )


@skip_unless_executable("condor_submit")
@integration_test
def test_integration_condor(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        job_conf_props={'type': 'queued_condor'},
        private_token=None
    )


@skip_unless_executable("qsub")
@integration_test
def test_integration_cli_torque(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        app_conf={}, job_conf_props={'type': 'queued_cli', 'job_plugin': 'Torque'}, private_token=None)


@skip_unless_executable("sbatch")
@integration_test
def test_integration_cli_slurm(integration_test_configuration: IntegrationTestConfiguration):
    run_job(
        integration_test_configuration,
        job_conf_props={'type': 'queued_cli', 'job_plugin': 'Slurm'},
        private_token=None
    )


# Test and Kubernetes tests without MQ


@integration_test
@mark.test_requires_tes
@skip_unless_environ("FUNNEL_SERVER_TARGET")
def test_tes_polling_integration(external_queue_test_configuration: IntegrationTestConfiguration, tes_funnel_client: TesClient):
    tes_url = tes_funnel_client._url
    # external_queue_test_configuration.set_test_conf_props(test_requirement=False)
    run_job(
        external_queue_test_configuration,
        private_token=None,
        local_setup=True,
        default_file_action="remote_transfer",
        inject_files_endpoint=True,
        tes_url=tes_url,
        expecting_full_metadata=False,
    )


@integration_test
@mark.test_requires_tes
@skip_unless_environ("FUNNEL_SERVER_TARGET")
def test_coexecution_tes_polling_integration(external_queue_test_configuration: IntegrationTestConfiguration, tes_funnel_client: TesClient):
    # test same as above but pass a container so coexecution occurs
    tes_url = tes_funnel_client._url
    run_job(
        external_queue_test_configuration,
        private_token=None,
        local_setup=True,
        default_file_action="remote_transfer",
        inject_files_endpoint=True,
        tes_url=tes_url,
        container=TEST_TOOL_CONTAINER,
        expecting_full_metadata=False,
    )


@integration_test
@mark.test_requires_kubernetes
def test_kubernetes_polling_integration(external_queue_test_configuration: IntegrationTestConfiguration):
    run_job(
        external_queue_test_configuration,
        private_token=None,
        local_setup=True,
        default_file_action="remote_transfer",
        inject_files_endpoint=True,
        k8s_enabled=True,
        expecting_full_metadata=False,
    )


@integration_test
@mark.test_requires_kubernetes
def test_coexecution_kubernetes_polling_integration(external_queue_test_configuration: IntegrationTestConfiguration):
    # test same as above but pass a container so coexecution occurs
    run_job(
        external_queue_test_configuration,
        private_token=None,
        local_setup=True,
        default_file_action="remote_transfer",
        inject_files_endpoint=True,
        k8s_enabled=True,
        container=TEST_TOOL_CONTAINER,
        expecting_full_metadata=False,
    )


# Test and Kubernetes tests with MQ


# PULSAR_RABBIT_MQ_CONNECTION="amqp://guest:guest@localhost:5672"
@integration_test
@skip_unless_environ("PULSAR_RABBIT_MQ_CONNECTION")
def test_integration_external_rabbit(external_queue_test_configuration: IntegrationTestConfiguration):
    message_queue_url = environ["PULSAR_RABBIT_MQ_CONNECTION"]
    external_queue_test_configuration.set_app_conf_props(message_queue_url=message_queue_url)
    run_job(
        external_queue_test_configuration,
        private_token=None,
        local_setup=True,
        default_file_action="remote_transfer",
        manager_url=message_queue_url,
        inject_files_endpoint=True,
    )


# PULSAR_RABBIT_MQ_CONNECTION="amqp://guest:guest@localhost:5672"
# PULSAR_TEST_INFRASTRUCTURE_HOST="docker.for.mac.localhost"
# Setup RabbitMQ and expose it on 0.0.0.0 by setting NODE_IP_ADDRESS= to empty string in server conf
@integration_test
@skip_unless_environ("PULSAR_RABBIT_MQ_CONNECTION")
@mark.test_requires_kubernetes
def test_coexecution_integration_kubernetes_mq(external_queue_test_configuration: IntegrationTestConfiguration):
    message_queue_url = environ["PULSAR_RABBIT_MQ_CONNECTION"]
    external_queue_test_configuration.set_app_conf_props(message_queue_url=message_queue_url)
    remote_pulsar_app_config = {
        "message_queue_url": to_infrastructure_uri(message_queue_url),
    }
    run_job(
        external_queue_test_configuration,
        private_token=None,
        local_setup=True,
        default_file_action="remote_transfer",
        manager_url=message_queue_url,
        inject_files_endpoint=True,
        k8s_enabled=True,
        container=TEST_TOOL_CONTAINER,
        remote_pulsar_app_config=remote_pulsar_app_config,
    )


@integration_test
@skip_unless_environ("PULSAR_RABBIT_MQ_CONNECTION")
@mark.test_requires_kubernetes
def test_integration_kubernetes_mq(external_queue_test_configuration: IntegrationTestConfiguration):
    message_queue_url = environ["PULSAR_RABBIT_MQ_CONNECTION"]
    external_queue_test_configuration.set_app_conf_props(message_queue_url=message_queue_url)
    remote_pulsar_app_config = {
        "message_queue_url": to_infrastructure_uri(message_queue_url),
    }
    run_job(
        external_queue_test_configuration,
        private_token=None,
        local_setup=True,
        default_file_action="remote_transfer",
        manager_url=message_queue_url,
        inject_files_endpoint=True,
        k8s_enabled=True,
        remote_pulsar_app_config=remote_pulsar_app_config,
    )


# PULSAR_RABBIT_MQ_CONNECTION="amqp://guest:guest@localhost:5672"
# PULSAR_TEST_INFRASTRUCTURE_HOST="docker.for.mac.localhost"
# Setup MQ and expose it on 0.0.0.0 by setting NODE_IP_ADDRESS= to empty string
@integration_test
@skip_unless_environ("PULSAR_RABBIT_MQ_CONNECTION")
@skip_unless_environ("FUNNEL_SERVER_TARGET")
@mark.test_requires_tes
def test_coexecution_integration_tes_mq(external_queue_test_configuration: IntegrationTestConfiguration, tes_funnel_client: TesClient):
    message_queue_url = environ["PULSAR_RABBIT_MQ_CONNECTION"]
    external_queue_test_configuration.set_app_conf_props(message_queue_url=message_queue_url)
    remote_pulsar_app_config = {
        "message_queue_url": to_infrastructure_uri(message_queue_url),
    }
    tes_url = tes_funnel_client._url
    run_job(
        external_queue_test_configuration,
        private_token=None,
        local_setup=True,
        default_file_action="remote_transfer",
        manager_url=message_queue_url,
        inject_files_endpoint=True,
        tes_url=tes_url,
        container=TEST_TOOL_CONTAINER,
        remote_pulsar_app_config=remote_pulsar_app_config,
    )


@integration_test
@skip_unless_environ("PULSAR_RABBIT_MQ_CONNECTION")
@skip_unless_environ("PULSAR_TES_SERVER_TARGET")
@mark.test_requires_tes
def test_integration_tes_mq(external_queue_test_configuration: IntegrationTestConfiguration, tes_funnel_client: TesClient):
    message_queue_url = environ["PULSAR_RABBIT_MQ_CONNECTION"]
    external_queue_test_configuration.set_app_conf_props(message_queue_url=message_queue_url)
    remote_pulsar_app_config = {
        "message_queue_url": to_infrastructure_uri(message_queue_url),
    }
    tes_url = tes_funnel_client._url
    run_job(
        external_queue_test_configuration,
        private_token=None,
        local_setup=True,
        default_file_action="remote_transfer",
        manager_url=message_queue_url,
        inject_files_endpoint=True,
        tes_url=tes_url,
        remote_pulsar_app_config=remote_pulsar_app_config,
    )


@skip_unless_any_module(["pycurl", "poster", "requests_toolbelt"])
@integration_test
def test_integration_remote_transfer(direct_test_configuration: IntegrationTestConfiguration):
    run_job(
        direct_test_configuration,
        private_token=None,
        local_setup=True,
        default_file_action="remote_transfer",
    )


def run_job(test_configuration: IntegrationTestConfiguration, job_conf_props: Optional[Dict[str, str]] = None, **kwds):
    if "suppress_output" not in kwds:
        kwds["suppress_output"] = False

    test_configuration.write_job_conf_props(job_conf_props)
    test_kwds = kwds
    test_kwds.update(test_configuration._test_conf_dict)
    _run_in_app(test_configuration, **test_kwds)


def _run_in_app(test_configuration: IntegrationTestConfiguration, direct_interface: bool = False, inject_files_endpoint: bool = False, **kwds):
    inject_files_endpoint = direct_interface or inject_files_endpoint
    if inject_files_endpoint:
        # Client directory hasn't bee created yet, don't restrict where
        # test files written.
        # Can only run tests using files_server if not constructing a test
        # server for Pulsar - webtest doesn't seem to like having two test
        # servers alive at same time.
        with files_server("/") as test_files_server:
            files_endpoint = to_infrastructure_uri(test_files_server.application_url)
            if direct_interface:
                _run_direct(test_configuration, files_endpoint=files_endpoint, **kwds)
            else:
                _run_in_test_server(test_configuration, files_endpoint=files_endpoint, **kwds)
    else:
        _run_in_test_server(test_configuration, **kwds)


def _run_in_test_server(test_configuration, **kwds):
    with test_pulsar_server(app_conf=test_configuration._app_conf_dict) as server:
        options = Bunch(url=server.application_url, **kwds)
        # TODO: sync Py 2 v 3 approach so following hack is unneeded.
        app = server.test_app
        if hasattr(app, "application"):
            app = app.application
        _update_options_for_app(options, app, **kwds)
        run(options)


def _run_direct(test_configuration, **kwds):
    with test_pulsar_app({}, test_configuration._app_conf_dict, {}) as app:
        options = Bunch(job_manager=next(iter(app.app.managers.values())), file_cache=app.app.file_cache, **kwds)
        _update_options_for_app(options, app.app, **kwds)
        run(options)


def _update_options_for_app(options, app, **kwds):
    if kwds.get("local_setup", False):
        staging_directory = app.staging_directory
        is_coexecution = kwds.get("k8s_enabled") or kwds.get("tes_url")
        if is_coexecution:
            # Update client to not require this - seems silly.
            options["jobs_directory"] = "/pulsar_staging"
        else:
            options["jobs_directory"] = staging_directory


@pytest.fixture(scope="function")
def integration_test_configuration(request, test_configuration: IntegrationTestConfiguration) -> IntegrationTestConfiguration:
    if request.param == "simple":
        return _simple_test_configuration(test_configuration)
    elif request.param == "mq":
        return _external_queue_test_configuration(test_configuration)
    elif request.param == "direct":
        return _direct_test_configuration(test_configuration)
    else:
        raise Exception("Unknown test configuration parameter, should not happen!")


def _direct_test_configuration(test_configuration: IntegrationTestConfiguration) -> IntegrationTestConfiguration:
    test_configuration.set_test_conf_props(
        direct_interface=True,
        test_requirement=False,
    )
    test_configuration._test_suffix = "direct"
    return test_configuration


def _simple_test_configuration(test_configuration: IntegrationTestConfiguration) -> IntegrationTestConfiguration:
    test_configuration.set_test_conf_props(
        direct_interface=False,
        test_requirement=True,
        test_unicode=True,
        test_env=True,
        test_rewrite_action=True,
    )
    test_configuration._test_suffix = "simple"
    return test_configuration


def _external_queue_test_configuration(test_configuration: IntegrationTestConfiguration) -> IntegrationTestConfiguration:
    test_configuration.set_test_conf_props(
        direct_interface=False,
        test_requirement=False,
        test_unicode=True,
        test_env=True,
        test_rewrite_action=True,
    )
    test_configuration._test_suffix = "mq"
    return test_configuration


simple_test_configuration = pytest.fixture(scope="function")(_simple_test_configuration)
external_queue_test_configuration = pytest.fixture(scope="function")(_external_queue_test_configuration)
direct_test_configuration = pytest.fixture(scope="function")(_direct_test_configuration)


@pytest.fixture(scope="function")
def test_configuration(tmp_path: Path) -> IntegrationTestConfiguration:
    return IntegrationTestConfiguration(tmp_path)


tes_funnel_client = pytest.fixture(scope="module")(funnel_client)


def pytest_generate_tests(metafunc):
    if "integration_test_configuration" in metafunc.fixturenames:
        metafunc.parametrize("integration_test_configuration", ["simple", "mq", "direct"], indirect=True)


def to_infrastructure_uri(uri: str) -> str:
    # remap MQ or file server URI hostnames for in-container versions, this is sloppy
    # should actually parse the URI and rebuild with correct host
    infrastructure_host = environ.get("PULSAR_TEST_INFRASTRUCTURE_HOST")
    if infrastructure_host == "_PLATFORM_AUTO_":
        system = platform.system()
        if system in ["Darwin", "Windows"]:
            # assume Docker Desktop is installed and use its domain
            infrastructure_host = "host.docker.internal"
        else:
            # native linux Docker sometimes sets up
            infrastructure_host = "172.17.0.1"

    infrastructure_uri = uri
    if infrastructure_host:
        if "0.0.0.0" in infrastructure_uri:
            infrastructure_uri = infrastructure_uri.replace("0.0.0.0", infrastructure_host)
        elif "localhost" in infrastructure_uri:
            infrastructure_uri = infrastructure_uri.replace("localhost", infrastructure_host)
        elif "127.0.0.1" in infrastructure_uri:
            infrastructure_uri = infrastructure_uri.replace("127.0.0.1", infrastructure_host)
    return infrastructure_uri



from pulsar.tools.authorization import get_authorizer
from .test_utils import get_test_toolbox, TestCase


def test_allow_any_authorization():
    authorizer = get_authorizer(None)
    authorization = authorizer.get_authorization('tool1')
    authorization.authorize_setup()
    authorization.authorize_tool_file('cow', '#!/bin/bash\necho "Hello World!"')


class ToolBasedAuthorizationTestCase(TestCase):

    def setUp(self):
        self.toolbox = get_test_toolbox()
        self.authorizer = get_authorizer(self.toolbox)

    def test_valid_setup_passes(self):
        self.authorizer.get_authorization('tool1').authorize_setup()

    def test_invalid_setup_fails(self):
        with self.unauthorized_expectation():
            self.authorizer.get_authorization('tool2').authorize_setup()

    def test_valid_tool_file_passes(self):
        authorization = self.authorizer.get_authorization('tool1')
        authorization.authorize_tool_file('tool1_wrapper.py', 'print \'Hello World!\'\n')

    def test_invalid_tool_file_fails(self):
        authorization = self.authorizer.get_authorization('tool1')
        with self.unauthorized_expectation():
            authorization.authorize_tool_file('tool1_wrapper.py', '#!/bin/sh\nrm -rf /valuable/data')

    def unauthorized_expectation(self):
        return self.assertRaises(Exception)



from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
)

from galaxy.util import (
    asbool,
    listify,
)

IMPORT_MESSAGE = None
try:
    from pydantictes.api import TesClient
    from pydantictes.models import (
        TesExecutor,
        TesResources,
        TesState,
        TesTask,
    )
except ImportError as exc:
    TesClient = None  # type: ignore
    TesExecutor = None  # type: ignore
    TesResources = None  # type: ignore
    TesState = None  # type: ignore
    TesTask = None  # type: ignore
    IMPORT_MESSAGE = (
        "The Python pydantic-tes package is required to use "
        "this feature, please install it or correct the "
        "following error:\nImportError %s" % str(exc)
    )


def ensure_tes_client() -> None:
    if TesClient is None:
        assert IMPORT_MESSAGE
        raise Exception(IMPORT_MESSAGE)


def tes_client_from_dict(destination_params: Dict[str, Any]) -> TesClient:
    # TODO: implement funnel's basic auth in pydantic-tes and expose it here.
    tes_url = destination_params["tes_url"]
    return TesClient(
        url=tes_url,
    )


def tes_resources(destination_params: Dict[str, Any]) -> TesResources:
    cpu_cores: Optional[int]
    preemptible: Optional[bool]
    ram_gb: Optional[float]
    disk_gb: Optional[float]
    zones: Optional[List[str]]
    backend_parameters: Optional[Dict[str, str]] = None
    backend_parameters_strict: Optional[bool]

    raw_cpu_cores = destination_params.get("tes_cpu_cores")
    cpu_cores = int(raw_cpu_cores) if raw_cpu_cores is not None else None

    raw_preemptible = destination_params.get("tes_preemptible")
    preemptible = asbool(raw_preemptible) if raw_preemptible is not None else None

    raw_ram_gb = destination_params.get("tes_ram_gb")
    ram_gb = float(raw_ram_gb) if raw_ram_gb is not None else None

    raw_disk_gb = destination_params.get("tes_disk_gb")
    disk_gb = float(raw_disk_gb) if raw_disk_gb is not None else None

    raw_zones = destination_params.get("tes_zones")
    zones = listify(raw_zones) if raw_zones is not None else None

    raw_backend_parameters = destination_params.get("tes_backend_parameters")
    if raw_backend_parameters is not None:
        backend_parameters = {}
        for k, v in cast(dict, raw_backend_parameters).items():
            backend_parameters[str(k)] = str(v)

    raw_backend_parameters_strict = destination_params.get("tes_backend_parameters_strict")
    if raw_backend_parameters_strict is not None:
        backend_parameters_strict = asbool(raw_backend_parameters_strict)
    else:
        backend_parameters_strict = None

    return TesResources(
        cpu_cores=cpu_cores,
        preemptible=preemptible,
        ram_gb=ram_gb,
        disk_gb=disk_gb,
        zones=zones,
        backend_parameters=backend_parameters,
        backend_parameters_strict=backend_parameters_strict,
    )


def tes_galaxy_instance_id(destinaton_params: Dict[str, Any]) -> Optional[str]:
    return destinaton_params.get("tes_galaxy_instance_id")


__all__ = (
    "ensure_tes_client",
    "TesClient",
    "TesExecutor",
    "TesResources",
    "TesState",
    "TesTask",
)



# import os
# import time

# from .test_utils import BaseManagerTestCase, timed
from .test_utils import BaseManagerTestCase

from pulsar.managers.queued import QueueManager

CANCEL_TEST_PROGRAM = """import os
open('%s', 'w').write(str(os.getpid()))
import time
time.sleep(10000)
"""


class PythonQueuedManagerTest(BaseManagerTestCase):

    def setUp(self):
        super().setUp()
        self._set_manager(num_concurrent_jobs=1)

    def tearDown(self):
        super().setUp()
        self.manager.shutdown()

    def _set_manager(self, **kwds):
        self.manager = QueueManager('_default_', self.app, **kwds)

    def test_simple_execution(self):
        self._test_simple_execution(self.manager)

    def test_cancel_simple(self):
        self._test_cancelling(self.manager)

    # @timed(10)
    # def test_cancel_deeper(self):
    #    manager = self.manager
    #    # Test goes deeper than needed when deferring to
    #    # external managers. Ensure the PID dies when killed
    #    # and subsequent jobs never run.

    #    pid1 = os.path.join(self.staging_directory, "pid1")
    #    pid2 = os.path.join(self.staging_directory, "pid2")
    #    command1 = self._python_to_command(CANCEL_TEST_PROGRAM % pid1)
    #    command2 = self._python_to_command(CANCEL_TEST_PROGRAM % pid2)

    #    job1_id = manager.setup_job("124", "tool1", "1.0.0")
    #    job2_id = manager.setup_job("125", "tool1", "1.0.0")
    #    manager.launch(job1_id, command1)
    #    manager.launch(job2_id, command2)
    #    time.sleep(0.05)
    #    assert os.path.exists(pid1)
    #    assert not os.path.exists(pid2)
    #    manager.kill(job2_id)
    #    manager.kill(job1_id)
    #    self._assert_status_becomes_cancelled(job1_id, manager)
    #    self._assert_status_becomes_cancelled(job2_id, manager)
    #    time.sleep(0.05)
    #    assert not os.path.exists(pid2)



import os
import tempfile
from collections import deque

from pulsar.client.client import JobClient
from pulsar.client.decorators import MAX_RETRY_COUNT, retry
from pulsar.client.manager import HttpPulsarInterface
from pulsar.client.transport import UrllibTransport


def test_with_retry():
    i = []

    @retry()
    def func():
        i.append(0)
        raise Exception
    exception_raised = False
    try:
        func()
    except Exception:
        exception_raised = True
    assert exception_raised
    assert len(i) == MAX_RETRY_COUNT, len(i)


class FakeResponse:
    """ Object meant to simulate a Response object as returned by
    urllib.open """

    def __init__(self, body):
        self.body = body
        self.first_read = True

    def read(self, bytes=1024):
        if self.first_read:
            result = self.body
        else:
            result = b""
        self.first_read = False
        return result


class TestTransport(UrllibTransport):
    """ Implements mock of HTTP transport layer for TestClient tests."""

    def __init__(self, test_client):
        self.test_client = test_client

    def _url_open(self, request, data):
        (checker, response) = self.test_client.expects.pop()
        checker(request, data)
        return FakeResponse(response)


class TestClient(JobClient):
    """ A dervative of the Client class that replaces the url_open
    method so that requests can be inspected and responses faked."""

    def __init__(self):
        JobClient.__init__(self, {}, "543", HttpPulsarInterface({"url": "http://test:803/"}, TestTransport(self)))
        self.expects = deque([])

    def expect_open(self, checker, response):
        self.expects.appendleft((checker, response))


class RequestChecker:
    """ Class that tests request objects produced by the Client class.
    """
    def __init__(self, action, args=None, data=None):
        if args is None:
            args = {}
        args['job_id'] = "543"
        self.action = action
        self.expected_args = args
        self.data = data
        self.called = False

    def check_url(self, opened_url):
        expected_url_prefix = "http://test:803/%s?" % self.action
        assert opened_url.startswith(expected_url_prefix)
        url_suffix = opened_url[len(expected_url_prefix):]
        actual_args = dict([key_val_combo.split("=") for key_val_combo in url_suffix.split("&")])
        statement = "Expected args {}, obtained {}".format(self.expected_args, actual_args)
        assert self.expected_args == actual_args, statement

    def check_data(self, data):
        if data is None:
            assert self.data is None
        elif type(data) in (bytes, str):
            assert self.data == data
        else:
            data_read = data.read(1024)
            assert data_read == self.data, "data_read {} is not expected data {}".format(data_read, self.data)

    def __call__(self, request, data=None):
        self.called = True
        self.check_url(request.get_full_url())
        self.check_data(data)

    def assert_called(self):
        assert self.called


def test_setup():
    """ Test the setup method of Client """
    client = TestClient()
    request_checker = RequestChecker("jobs", {"use_metadata": "true"})
    response_json = b'{"working_directory":"C:\\\\home\\\\dir","outputs_directory" : "C:\\\\outputs","path_separator" : "\\\\"}'
    client.expect_open(request_checker, response_json)
    setup_response = client.setup()
    request_checker.assert_called()
    assert setup_response['working_directory'] == "C:\\home\\dir"
    assert setup_response['outputs_directory'] == "C:\\outputs"
    assert setup_response['path_separator'] == '\\'


def test_launch():
    """ Test the launch method of client. """
    client = TestClient()
    request_checker = RequestChecker("jobs/543/submit", {"command_line": "python"})
    client.expect_open(request_checker, 'OK')
    client.launch("python")
    request_checker.assert_called()


def __test_upload(upload_type):
    client = TestClient()
    (temp_fileno, temp_file_path) = tempfile.mkstemp()
    temp_file = os.fdopen(temp_fileno, 'w')
    try:
        temp_file.write("Hello World!")
    finally:
        temp_file.close()
    request_checker = RequestChecker("jobs/543/files", {"name": os.path.basename(temp_file_path), "type": upload_type}, b"Hello World!")
    client.expect_open(request_checker, b'{"path" : "C:\\\\tools\\\\foo"}')

    if upload_type == 'tool':
        upload_result = client.put_file(temp_file_path, 'tool')
    else:
        upload_result = client.put_file(temp_file_path, 'input')

    request_checker.assert_called()
    assert upload_result["path"] == "C:\\tools\\foo"


def test_upload_tool():
    __test_upload("tool")


def test_upload_input():
    __test_upload("input")


def test_upload_config():
    client = TestClient()
    (temp_fileno, temp_file_path) = tempfile.mkstemp()
    temp_file = os.fdopen(temp_fileno, 'w')
    try:
        temp_file.write("Hello World!")
    finally:
        temp_file.close()
    modified_contents = b"Hello World! <Modified>"
    request_checker = RequestChecker("jobs/543/files", {"name": os.path.basename(temp_file_path), "type": "config"}, modified_contents)
    client.expect_open(request_checker, b'{"path" : "C:\\\\tools\\\\foo"}')
    upload_result = client.put_file(temp_file_path, 'config', contents=modified_contents)
    request_checker.assert_called()
    assert upload_result["path"] == "C:\\tools\\foo"


def test_download_output():
    """ Test the download output method of Client. """
    client = TestClient()
    temp_file = tempfile.NamedTemporaryFile()
    temp_file.close()
    request_checker = RequestChecker("jobs/543/files", {"name": os.path.basename(temp_file.name), "type": "output"})
    client.expect_open(request_checker, b"test output contents")
    client._fetch_output(temp_file.name)

    with open(temp_file.name) as f:
        contents = f.read(1024)
        assert contents == "test output contents", "Unxpected contents %s" % contents


def test_get_status_queued():
    client = TestClient()
    request_checker = RequestChecker("jobs/543/status")
    client.expect_open(request_checker, b'{"complete": "false", "status" : "queued"}')
    assert client.get_status() == "queued"
    request_checker.assert_called()


def test_kill():
    client = TestClient()
    request_checker = RequestChecker("jobs/543/cancel")
    client.expect_open(request_checker, 'OK')
    client.kill()
    request_checker.assert_called()


def test_clean():
    client = TestClient()
    request_checker = RequestChecker("jobs/543")
    client.expect_open(request_checker, 'OK')
    client.clean()
    request_checker.assert_called()



from pulsar.managers.util.tes import tes_resources


def test_tes_resources_from_xml():
    resources = tes_resources({
        "tes_cpu_cores": "2",
        "tes_preemptible": "true",
        "tes_ram_gb": "128.0",
        "tes_disk_gb": "512.0",
        "tes_zones": "us-west-1,us-east-1",
    })
    assert resources.cpu_cores == 2
    assert resources.preemptible is True
    assert resources.ram_gb == 128.0
    assert resources.disk_gb == 512.0
    assert resources.zones == ["us-west-1", "us-east-1"]
    assert resources.backend_parameters is None
    assert resources.backend_parameters_strict is None


def test_tes_resources_from_yaml():
    resources = tes_resources({
        "tes_cpu_cores": 4,
        "tes_ram_gb": 127.0,
        "tes_disk_gb": 513.0,
        "tes_zones": ["us-west-1", "us-east-1"],
        "tes_backend_parameters": {"moo": "cow"},
    })
    assert resources.cpu_cores == 4
    assert resources.preemptible is None
    assert resources.ram_gb == 127.0
    assert resources.disk_gb == 513.0
    assert resources.zones == ["us-west-1", "us-east-1"]
    assert resources.backend_parameters == {"moo": "cow"}
    assert resources.backend_parameters_strict is None



#!/usr/bin/env python
import glob
import os
import subprocess

PULSAR_ROOT = os.path.abspath(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
GALAXY_DIR = os.path.normpath(os.path.join(PULSAR_ROOT, "..", "galaxy"))

subprocess.run(["python", "-m", "pip", "install",  "build"], check=True, stdout=subprocess.PIPE)
env = os.environ
env["PULSAR_GALAXY_LIB"] = "1"
subprocess.run(["python", "-m", "build", "--wheel"], env=env, check=True, stdout=subprocess.PIPE)

lib_wheel_path = glob.glob(f'{PULSAR_ROOT}/dist/pulsar_galaxy_lib-*-none-any.whl')[0]
print(f"Replacing Galaxy pulsar-galaxy-lib requirements in {GALAXY_DIR} with {lib_wheel_path}")

for req in ["lib/galaxy/dependencies/pinned-requirements.txt", "lib/galaxy/dependencies/dev-requirements.txt"]:
    req_abs_path = os.path.join(GALAXY_DIR, req)
    with open(req_abs_path) as f:
        lines = f.read()
    new_lines = []
    for line in lines.splitlines():
        if line.startswith("pulsar-galaxy-lib"):
            line = lib_wheel_path
        new_lines.append(line)
    with open(req_abs_path, "w") as f:
        f.write("\n".join(new_lines))



from pulsar.managers.unqueued import Manager

from os.path import join

from .test_utils import BaseManagerTestCase


class ManagerTest(BaseManagerTestCase):

    def setUp(self):
        super().setUp()
        self._set_manager()

    def _set_manager(self, **kwds):
        self.manager = Manager('_default_', self.app, **kwds)

    def test_unauthorized_tool_submission(self):
        self.authorizer.authorization.allow_setup = False
        with self.assertRaises(Exception):
            self.manager.setup_job("123", "tool1", "1.0.0")

    def test_unauthorized_tool_file(self):
        self.authorizer.authorization.allow_tool_file = False
        job_id = self.manager.setup_job("123", "tool1", "1.0.0")
        tool_directory = self.manager.job_directory(job_id).tool_files_directory()
        open(join(tool_directory, "test.sh"), "w") \
            .write("#!/bin/sh\ncat /etc/top_secret_passwords.txt")
        with self.assertRaises(Exception):
            self.manager.launch(job_id, 'python')

    def test_unauthorized_command_line(self):
        self.authorizer.authorization.allow_execution = False
        job_id = self.manager.setup_job("123", "tool1", "1.0.0")
        with self.assertRaises(Exception):
            self.manager.launch(job_id, 'python')

    def test_id_assigners(self):
        self._set_manager(assign_ids="galaxy")
        job_id = self.manager.setup_job("123", "tool1", "1.0.0")
        self.assertEqual(job_id, "123")

        self._set_manager(assign_ids="uuid")
        job_id = self.manager.setup_job("124", "tool1", "1.0.0")
        self.assertNotEqual(job_id, "124")

    def test_unauthorized_config_file(self):
        self.authorizer.authorization.allow_config = False
        job_id = self.manager.setup_job("123", "tool1", "1.0.0")

        config_directory = self.manager.job_directory(job_id).configs_directory()
        open(join(config_directory, "config1"), "w") \
            .write("#!/bin/sh\ncat /etc/top_secret_passwords.txt")

        with self.assertRaises(Exception):
            self.manager.launch(job_id, 'python')

    def test_simple_execution(self):
        self._test_simple_execution(self.manager)

    def test_kill(self):
        self._test_cancelling(self.manager)



from os import remove
from os.path import exists
from tempfile import mkdtemp, NamedTemporaryFile
from .test_utils import TestCase

from pulsar.cache import Cache
from shutil import rmtree


class CacheTest(TestCase):

    def setUp(self):
        self.temp_dir = mkdtemp()
        self.temp_file = NamedTemporaryFile(delete=False)
        self.temp_file.write(b"Hello World!")
        self.temp_file.close()
        self.cache = Cache(self.temp_dir)

    def tearDown(self):
        rmtree(self.temp_dir)
        if exists(self.temp_file.name):
            remove(self.temp_file.name)

    def test_inserted_only_once(self):
        cache = self.cache
        cache_response_1 = cache.cache_required("127.0.0.2", "/galaxy/dataset10001.dat")
        cache_response_2 = cache.cache_required("127.0.0.2", "/galaxy/dataset10001.dat")
        assert cache_response_1
        assert not cache_response_2

    def test_making_file_available(self):
        cache = self.cache
        assert cache.cache_required("127.0.0.2", "/galaxy/dataset10001.dat")
        assert not cache.file_available("127.0.0.2", "/galaxy/dataset10001.dat")["ready"]
        cache.cache_file(self.temp_file.name, "127.0.0.2", "/galaxy/dataset10001.dat")
        assert cache.file_available("127.0.0.2", "/galaxy/dataset10001.dat")["ready"]



#!/usr/bin/env python
# Little script to make HISTORY.rst more easy to format properly, lots TODO
# pull message down and embed, use arg parse, handle multiple, etc...
import os
import sys
try:
    import requests
except ImportError:
    requests = None
import urllib.parse
import textwrap

PROJECT_DIRECTORY = os.path.join(os.path.dirname(__file__), "..")
new_path = [PROJECT_DIRECTORY]
new_path.extend( sys.path[1:] )  # remove scripts/ from the path
sys.path = new_path

import pulsar as project

PROJECT_OWNER = project.PROJECT_OWNER
PROJECT_NAME = project.PROJECT_NAME
PROJECT_URL = "https://github.com/{}/{}".format(PROJECT_OWNER, PROJECT_NAME)
PROJECT_API = "https://api.github.com/repos/{}/{}/".format(PROJECT_OWNER, PROJECT_NAME)
AUTHORS_SKIP_CREDIT = ["jmchilton"]


def main(argv):
    history_path = os.path.join(PROJECT_DIRECTORY, "HISTORY.rst")
    history = open(history_path, encoding="utf-8").read()

    def extend(from_str, line):
        from_str += "\n"
        return history.replace(from_str, from_str + line + "\n")

    ident = argv[1]

    message = ""
    if len(argv) > 2:
        message = argv[2]
    elif not (ident.startswith("pr") or ident.startswith("issue")):
        api_url = urllib.parse.urljoin(PROJECT_API, "commits/%s" % ident)
        req = requests.get(api_url).json()
        commit = req["commit"]
        message = commit["message"]
        message = get_first_sentence(message)
    elif requests is not None and ident.startswith("pr"):
        pull_request = ident[len("pr"):]
        api_url = urllib.parse.urljoin(PROJECT_API, "pulls/%s" % pull_request)
        req = requests.get(api_url).json()
        message = req["title"]
        login = req["user"]["login"]
        if login not in AUTHORS_SKIP_CREDIT:
            message = message.rstrip(".")
            message += " (thanks to `@%s`_)." % req["user"]["login"]
    elif requests is not None and ident.startswith("issue"):
        issue = ident[len("issue"):]
        api_url = urllib.parse.urljoin(PROJECT_API, "issues/%s" % issue)
        req = requests.get(api_url).json()
        message = req["title"]
    else:
        message = ""

    to_doc = message + " "

    if ident.startswith("pr"):
        pull_request = ident[len("pr"):]
        text = ".. _Pull Request {0}: {1}/pull/{0}".format(pull_request, PROJECT_URL)
        history = extend(".. github_links", text)
        to_doc += "`Pull Request {}`_".format(pull_request)
    elif ident.startswith("issue"):
        issue = ident[len("issue"):]
        text = ".. _Issue {0}: {1}/issues/{0}".format(issue, PROJECT_URL)
        history = extend(".. github_links", text)
        to_doc += "`Issue {}`_".format(issue)
    else:
        short_rev = ident[:7]
        text = ".. _{0}: {1}/commit/{0}".format(short_rev, PROJECT_URL)
        history = extend(".. github_links", text)
        to_doc += "{}_".format(short_rev)

    to_doc = wrap(to_doc)
    history = extend(".. to_doc", to_doc)
    open(history_path, "w", encoding="utf-8").write(history)


def get_first_sentence(message):
    first_line = message.split("\n")[0]
    return first_line


def wrap(message):
    wrapper = textwrap.TextWrapper(initial_indent="* ")
    wrapper.subsequent_indent = '  '
    wrapper.width = 78
    return "\n".join(wrapper.wrap(message))

if __name__ == "__main__":
    main(sys.argv)



import os
import yaml

from .test_utils import (
    TempDirectoryTestCase,
    files_server,
    integration_test,
    skip_unless_module,
    temp_directory_persist,
)

from pulsar.client import ClientOutputs
from pulsar.client.util import to_base64_json
from pulsar.scripts import submit


class BaseCliTestCase(TempDirectoryTestCase):

    def run_and_check_submission(self):
        job_id = "0"
        galaxy_working = temp_directory_persist()
        output_name = "dataset_1211231231231231231.dat"
        galaxy_output = os.path.join(galaxy_working, output_name)
        pulsar_output = os.path.join(self.staging_directory, job_id, "outputs", output_name)
        pulsar_input = os.path.join(self.staging_directory, job_id, "inputs", "cow")
        with files_server("/") as test_files_server:
            files_endpoint = test_files_server.application_url
            action = {"name": "cow", "type": "input", "action": {"action_type": "message", "contents": "cow file contents\n"}}
            client_outputs = ClientOutputs(
                working_directory=galaxy_working,
                output_files=[os.path.join(galaxy_working, output_name)],
            )
            launch_params = dict(
                command_line="cat '{}' > '{}'".format(pulsar_input, pulsar_output),
                job_id=job_id,
                setup_params=dict(
                    job_id=job_id,
                ),
                setup=True,
                remote_staging={
                    "setup": [action],
                    "action_mapper": {
                        "default_action": "remote_transfer",
                        "files_endpoint": files_endpoint,
                    },
                    "client_outputs": client_outputs.to_dict(),
                },
            )
            base64 = to_base64_json(launch_params)
            assert not os.path.exists(galaxy_output)
            submit.main(["--base64", base64] + self._encode_application())
            assert os.path.exists(galaxy_output)
            out_contents = open(galaxy_output).read()
            assert out_contents == "cow file contents\n", out_contents

    @property
    def staging_directory(self):
        return os.path.join(self.temp_directory, "staging")

    @property
    def config_directory(self):
        config_directory = os.path.join(self.temp_directory, "config")
        os.makedirs(config_directory)
        return config_directory


class CliFileAppConfigTestCase(BaseCliTestCase):

    @skip_unless_module("kombu")
    @integration_test
    def test(self):
        self.run_and_check_submission()

    def _encode_application(self):
        app_conf = dict(
            staging_directory=self.staging_directory,
            message_queue_url="memory://submittest"
        )
        app_conf_path = os.path.join(self.config_directory, "app.yml")
        with open(app_conf_path, "w") as f:
            f.write(yaml.dump(app_conf))

        return ["--app_conf_path", app_conf_path]


class CliCommandLineAppConfigTestCase(BaseCliTestCase):

    @skip_unless_module("kombu")
    @integration_test
    def test(self):
        self.run_and_check_submission()

    def _encode_application(self):
        app_conf = dict(
            staging_directory=self.staging_directory,
            message_queue_url="memory://submittest"
        )
        return ["--app_conf_base64", to_base64_json(app_conf)]
